--[[
This module is only loaded when 'debug' is given as a run argument.
In that case, to access the editor in-app, call:
	B(level_id, tileset [, w, h])
Editor controls:
	F7: Export level.
	F8: Test Level (will also return you to the editor if hit while testing).
	Q: -- Hold and move mouse to reduce level size.
	W: ++ Hold and move mouse to increase level size.
	LCtrl: Hold and move mouse to move the camera.
	LMB: Place Tile.
	RMB: Remove Tile.
		Hold LShift with the last two to select and mass-change tiles.
	MMB: Pick-Tile, drag to select and copy tiles.
	V: Paste tiles. (Mouse = Top-left of selection) Places upon release, hold to show area.
	NonusBackslash (Right of LShift): Hold and move mouse up/down to change the current tile.
	D: Place the starting point.
	E: Place the finish.
	S: Place an Object/Entity. You may need to edit the output for more advanced stuff.
	1: Toggle Entity <> Object
	2+3: Change current Entity/Object to place (displayed top-left)
Non given level keys:
	collect: Whether or not you can collect stuff here (bool or number).
	out: Tile-ID to use outside level range.
	extend: Extend the level's border-tiles into infinity.
]]

local love = love
local ceil, floor = math.ceil, math.floor
local pairs, ipairs = pairs, ipairs
local window, input, colors = window, input, colors

local build = {}

local color = {
	white   = colors.vec4(1, 1, 1),
	bg      = colors.vec4(0, 0.25, 0.5),
	marker  = colors.vec4(1, 1, 0, 0.5),
	outline = colors.vec4(1, 0, 1, 0.5),
	select  = colors.vec4(0, 1, 1, 0.5)
}

local active = false
local initialized = false

local tiles, level
local scroll, select
local tile, ent, id, finish, start, object
local useEntity

local objectTypes = {
	"Barrier", "CheckPoint", "Finish", "Platform",
	"SceneTrigger", "Switch", "WarpDoor"
}
local objectArgs = {
	Barrier = [=[true --[[hori]], 3, nil]=],
	Platform = [=[3, 3, false --[[left]], false --[[vert]]]=],
	SceneTrigger = [[3, 2, "the_end", false]],
	Switch = [=["up" --[[, Object, ...]]]=],
	WarpDoor = [["test"]]
}
local entityTypes = love.filesystem.getDirectoryItems("jnr/entity")
for k,v in pairs(entityTypes) do entityTypes[k] = v:gsub("%.lua$", "") end

local lastfile = love.filesystem.read(".lastbuild.txt") or "##build"

local function savelevel(path)
	if path then
		path = "jnr/build_out/"..path..".lua"
	else
		path = "jnr/level/##build.lua"
	end

	love.filesystem.write(path, "return {\n\tlevel = {"..(function()
			local s = ""
			for y,v in ipairs(level) do
				s = s.."{"
				for x,v in ipairs(v) do
					s = s..tostring(v)..","
				end
				s = s.."},"
			end
			return s
		end)().."},\n\ttiles = \""..(tiles.id)..
		"\",\n\tspawn = {"..tostring(start[1])..", "..tostring(start[2]).."},\n\t"..
		(finish and ("finish = {"..tostring(finish[1])..", "..tostring(finish[2]).."}") or "finish = nil")..
		",\n\tobject = {"..(function()
			if #object > 0 then
				local s = "\n"
				for y,v in pairs(object) do
					if v[1] == "Entity" then
						s = s.."\t\t{'"..tostring(v[1]).."', "..tostring(v[2])..", "..tostring(v[3])..", '"..tostring(v[4]).."'},\n"
					elseif objectArgs[v[1]] then
						s = s.."\t\t{'"..tostring(v[1]).."', "..tostring(v[2])..", "..tostring(v[3])..", "..objectArgs[v[1]].."},\n"
					else
						s = s.."\t\t{'"..tostring(v[1]).."', "..tostring(v[2])..", "..tostring(v[3]).."},\n"
					end
				end
				return s.."\t"
			else
				return ""
			end
		end)().."}\n}")
end

function build.deinit()
	love.mouse.setVisible(false)
end

function build.init(path, tileset, w, h)
	fileT.require("jnr/level/#names.txt")["##build"] = "Build Mode!"

	love.mouse.setVisible(true)
	initialized = true
	if path == nil then path = lastfile end
	love.filesystem.write(".lastbuild.txt", path)

	id = path
	start = {1, 1}

	love.filesystem.createDirectory("jnr/level")
	love.filesystem.createDirectory("jnr/build_out")

	input.bindKey("f8", "toggleBuild")
	input.bindEvent(function()
		if active then
			save.set(nil)
			save.reset()
			save.data.hub = "##build"

			savelevel()
			game.setState("jnr", "##build")
			active = false
			love.mouse.setVisible(false)
		else
			game.setState("build", game.noInit)
			active = true
			love.mouse.setVisible(true)
		end
	end, "toggleBuild")

	local leveldata
	if path then
		path = "jnr/build_out/"..path..".lua"
	else
		path = "jnr/level/##build.lua"
	end
	if love.filesystem.isFile(path) then
		leveldata = love.filesystem.load(path)()
		finish = leveldata.finish
		start  = leveldata.spawn or {1, 1}
		object = leveldata.object or {}
	else
		if w == nil then w = 1 end
		if h == nil then h = 1 end

		local data = {}
		for y=1, h do data[y] = {} for x=1, w do data[y][x] = 0 end end
		leveldata = {
			level = data,
			tiles = tileset
		}
		object = {}
	end
	local tiledata = love.filesystem.load("jnr/tiles/"..(leveldata.tiles or "forest")..".lua")()

	local tileImage = assets.image("assets/tiles/"..tiledata.image..".png")
	tiles = {
		ts = tiledata.ts or 9,
		id = leveldata.tiles or "forest",
		image = tileImage
	}
	tiles.batch = love.graphics.newSpriteBatch(tileImage, 10000, "stream")

	tiles.quad = {}
	for k,v in ipairs(tiledata.data) do
		tiles.quad[#tiles.quad+1] = love.graphics.newQuad(v[1]*tiles.ts-tiles.ts, v[2]*tiles.ts-tiles.ts, tiles.ts, tiles.ts, tileImage:getDimensions())
	end

	level = leveldata.level

	scroll = {
		x = 0, y = 0
	}
	select = {
		x = 1, y = 1,
		w = 1, h = 1
	}

	input.bindKey("lctrl", "BScroll")
	input.bindKey("lshift", "BSelect")
	input.bindKey("nonusbackslash", "BTile")
	input.bindKey("q", "BReduce")
	input.bindKey("w", "BExpand")
	input.bindKey("e", "BFinish")
	input.bindKey("d", "BStart")
	input.bindKey("s", "BMarker")
	input.bindKey("v", "BPaste")

	input.bindKey("1", "BToggleEnt")
	input.bindKey("2", "BDecreaseEnt")
	input.bindKey("3", "BIncreaseEnt")
	input.bindEvent(function()
		useEntity = not useEntity
		ent = 1
	end, "BToggleEnt")
	input.bindEvent(function()
		ent = ent + 1
		if useEntity then
			if ent > #entityTypes then ent = 1 end
		else
			if ent > #objectTypes then ent = 1 end
		end
	end, "BIncreaseEnt")
	input.bindEvent(function()
		ent = ent - 1
		if useEntity then
			if ent < 1 then ent = #entityTypes end
		else
			if ent < 1 then ent = #objectTypes end
		end
	end, "BDecreaseEnt")

	input.bindKey("f7", "export")
	input.bindEvent(function()
		savelevel(id)
		love.system.openURL("file:///"..love.filesystem.getSaveDirectory().."/jnr/build_out")
	end, "export")

	tile = 1
	ent = 1
	useEntity = false

	active = true
end

local orig
local function tileSim(tile, tx, ty)
	if input.down("BSelect") or tile == "copy" then
		if not orig then
			orig = {
				x = select.x,
				y = select.y,
				tile = tile
			}
		else
			if tx > orig.x then
				select.x = orig.x
				select.w = tx - orig.x + 1
			elseif tx < orig.x then
				select.x = tx
				select.w = orig.x - tx + 1
			else
				select.x = tx
				select.w = 1
			end

			if ty > orig.y then
				select.y = orig.y
				select.h = ty - orig.y + 1
			elseif ty < orig.y then
				select.y = ty
				select.h = orig.y - ty + 1
			else
				select.y = ty
				select.h = 1
			end
		end
	elseif level[ty] then
		if level[ty][tx] then
			level[ty][tx] = tile
		end
	end
end

local function clonetable(t)
	local n = {}
	if t then
		for k,v in pairs(t) do
			n[k] = v
		end
	end
	return n
end

local lx, ly = 0, 0
local ltx, lty = 0, 0
local paste
function build.update(dt)
	local x = (love.mouse.getX() / window.s + scroll.x - window.vw * 0.5)
	local y = (love.mouse.getY() / window.s + scroll.y - window.vh * 0.5)

	local tx = ceil(x / tiles.ts)
	local ty = ceil(y / tiles.ts)

	if input.down("BScroll") then
		scroll.x = scroll.x + (lx - love.mouse.getX())/window.s
		scroll.y = scroll.y + (ly - love.mouse.getY())/window.s
	elseif input.down("BTile") then
		if ty < lty then
			tile = tile + 1
			if tile > #tiles.quad then tile = 1 end
		elseif ty > lty then
			tile = tile - 1
			if tile < 1 then tile = #tiles.quad end
		end
	elseif input.down("BExpand") then
		if tx > ltx then
			for i,v in ipairs(level) do
				table.insert(v, v[#v] or 0)
			end
		elseif tx < ltx then
			for i,v in ipairs(level) do
				table.insert(v, 1, v[1] or 0)
			end

			start[1] = start[1] + 1
			if finish then finish[1] = finish[1] + 1 end
			for k,v in pairs(object) do v[2] = v[2] + 1 end
		end
		if ty > lty then
			table.insert(level, clonetable(level[#level]))
		elseif ty < lty then
			table.insert(level, 1, clonetable(level[1]))

			start[2] = start[2] + 1
			if finish then finish[2] = finish[2] + 1 end
			for k,v in pairs(object) do v[3] = v[3] + 1 end
		end
	elseif input.down("BReduce") then
		if tx > ltx then
			for i,v in ipairs(level) do
				table.remove(v, #v)
			end
		elseif tx < ltx then
			for i,v in ipairs(level) do
				table.remove(v, 1)
			end

			start[1] = start[1] - 1
			if finish then finish[1] = finish[1] - 1 end
			for k,v in pairs(object) do v[2] = v[2] - 1 end
		end
		if ty > lty then
			table.remove(level, #level)
		elseif ty < lty then
			table.remove(level, 1)

			start[2] = start[2] - 1
			if finish then finish[2] = finish[2] - 1 end
			for k,v in pairs(object) do v[3] = v[3] - 1 end
		end
	elseif input.new("BPaste") then
		local num = {}
		for s in love.system.getClipboardText():gmatch("%d*") do
			num[#num+1] = tonumber(s)
		end

		paste = { w = num[1], h = num[2] }
		table.remove(num, 1) table.remove(num, 1)
		paste.val = num

		if not (paste.w and paste.h) then
			paste = nil
		end
	end

	if love.mouse.isDown(1) then
		tileSim(tile, tx, ty)
	elseif love.mouse.isDown(2) then
		tileSim(0, tx, ty)
	elseif love.mouse.isDown(3) then
		tileSim("copy", tx, ty)
	elseif input.new("BFinish") then
		if finish then
			if finish[1] ~= tx or finish[2] ~= ty then
				finish = {tx, ty}
			else
				finish = nil
			end
		else
			finish = {tx, ty}
		end
	elseif input.new("BStart") then
		if start then
			start = {tx, ty}
		end
	elseif input.new("BMarker") then
		local id
		for k,v in pairs(object) do
			if v[2] == tx and v[3] == ty then
				local objType = useEntity and "Entity" or objectTypes[ent]
				if v[1] == objType and (not useEntity or v[4] == entityTypes[ent]) then
					id = k
					object[k] = nil
				else
					v[1] = objType
				end
				break
			end
		end
		if not id then
			if useEntity then
				object[#object+1] = {"Entity", tx, ty, entityTypes[ent]}
			else
				object[#object+1] = {objectTypes[ent], tx, ty}
			end
		end
	end

	if orig then
		if not love.mouse.isDown(1, 2, 3) then
			if orig.tile == "copy" then
				if select.w == 1 and select.h == 1 then
					if level[select.y] then
						local t = level[select.y][select.x]
						if not (t == 0 or not t) then
							tile = t
						end
					end
				else
					local copy = ("%d %d"):format(select.w, select.h)

					for y=select.y, select.y+select.h-1 do
						if level[y] then
							for x=select.x, select.x+select.w-1 do
								copy = copy .. (" %d"):format(level[y][x] or 0)
							end
						end
					end

					love.system.setClipboardText(copy)
				end
			else
				for y=select.y, select.y+select.h-1 do
					if level[y] then
						for x=select.x, select.x+select.w-1 do
							if level[y][x] then
								level[y][x] = orig.tile
							end
						end
					end
				end
			end

			orig = nil
		end
	elseif paste and input.down("BPaste") then
		select.x = tx
		select.y = ty
		select.w = paste.w
		select.h = paste.h
	elseif paste then
		local i = 0

		for y=select.y, select.y+paste.h-1 do
			if level[y] then
				for x=select.x, select.x+paste.w-1 do
					i = i + 1
					if level[y][x] and paste.val[i] then
						level[y][x] = paste.val[i]
					end
				end
			end
		end

		paste = nil
	else
		select.x = tx
		select.y = ty
		select.w = 1
		select.h = 1
	end

	lx, ly = love.mouse.getPosition()
	ltx, lty = tx, ty
end

function build.draw()
	love.graphics.push()

	love.graphics.setColor(color.bg)
	love.graphics.rectangle("fill", 0, 0, window.w, window.h)
	love.graphics.setColor(color.white)

	love.graphics.scale(window.s)
	love.graphics.translate(floor(-scroll.x + window.vw*0.5), floor(-scroll.y + window.vh*0.5))

	tiles.batch:clear()
	for y=floor((scroll.y-window.vh*0.5)/tiles.ts), ceil((scroll.y+window.vh*0.5)/tiles.ts) do
		if level[y] then
			for x=floor((scroll.x-window.vw*0.5)/tiles.ts), ceil((scroll.x+window.vw*0.5)/tiles.ts) do
				local v = level[y][x]
				local q = tiles.quad[v]
				if q then
					tiles.batch:add(q, x*tiles.ts-tiles.ts, y*tiles.ts-tiles.ts)
				end
			end
		end
	end
	love.graphics.draw(tiles.batch)

	love.graphics.setColor(color.marker)
	for k,v in pairs(object) do
		local x = v[2]*tiles.ts-tiles.ts*0.5-2
		local y = v[3]*tiles.ts-tiles.ts*0.5-2
		love.graphics.polygon("line", x, y, x+4, y, x+4, y+4, x, y+4)
		love.graphics.print(v[1]..(v[1] == "Entity" and ", "..v[4] or ""), x+4, y)
	end

	if level[1] then
		love.graphics.setColor(color.outline)
		love.graphics.rectangle("line", 0, 0, tiles.ts*#level[1], tiles.ts*#level)
	end

	if start then
		local x = start[1]*tiles.ts-tiles.ts*0.5-2
		local y = start[2]*tiles.ts-tiles.ts*0.5-2
		love.graphics.polygon("line", x, y, x+4, y, x+4, y+4, x, y+4)
	end

	love.graphics.setColor(color.select)
	love.graphics.rectangle("line", select.x*tiles.ts-tiles.ts, select.y*tiles.ts-tiles.ts, select.w*tiles.ts, select.h*tiles.ts)

	if finish then
		local x = finish[1]*tiles.ts-tiles.ts*0.5
		local y = finish[2]*tiles.ts-tiles.ts*0.5
		love.graphics.polygon("line", x+4, y, x, y+4, x-4, y, x, y-4)
	end

	love.graphics.pop()

	if input.down("BTile") then
		love.graphics.setColor(color.marker)
		love.graphics.polygon("line", window.w*0.5-tiles.ts*window.s, window.h*0.5, window.w*0.5, window.h*0.5-tiles.ts*window.s, window.w*0.5+tiles.ts*window.s, window.h*0.5, window.w*0.5, window.h*0.5+tiles.ts*window.s)

		love.graphics.setColor(color.white)

		local y = floor(window.h*0.5-(tiles.ts + 1) * 3 * window.s)
		for i=tile-3, tile+3 do
			if i < 1 then i = #tiles.quad + i
			elseif i > #tiles.quad then i = i - #tiles.quad end
			love.graphics.draw(tiles.image, tiles.quad[i], floor(window.w*0.5), y, 0, window.s, nil, ceil(tiles.ts*0.5), ceil(tiles.ts*0.5))

			y = y + (tiles.ts + 1) * window.s
		end
	else
		love.graphics.setColor(color.marker)
		local points = {window.w*0.5+4, window.h*0.5, window.w*0.5, window.h*0.5+4, window.w*0.5-4, window.h*0.5, window.w*0.5, window.h*0.5-4}
		if input.down("BExpand") then
			love.graphics.line(points[1], points[2], points[5], points[6])
			love.graphics.line(points[3], points[4], points[7], points[8])
		elseif input.down("BReduce") then
			love.graphics.line(points[1], points[2], points[5], points[6])
		else
			love.graphics.polygon("line", points)
		end
		love.graphics.setColor(color.white)

		love.graphics.print(("Selection: XY %d, %d :: WH %d, %d\n%s: %s"):format(select.x, select.y, select.w, select.h, useEntity and "Entity" or "Object", useEntity and entityTypes[ent] or objectTypes[ent]), 1, 1)
		love.graphics.draw(tiles.image, tiles.quad[tile], 1, 27, 0, window.s)
	end
end

return build
