local love = love
local abs = math.abs
local require, pcall, ipairs, unpack, collectgarbage = require, pcall, ipairs, unpack, collectgarbage
local window, colors = window, colors
local demoV if _demo then demoV = "Demo! Version ".._gameVersion end

local log = newLog "[Game]"

local game = {}

game.noInit = "const: game.noInit"

game.states = {
	jnr   = require "jnr",
	menu  = require "menu",
	scene = require "scene",
	title = require "title",
	none = {
		init = function()end,
		update = function()end,
		draw = function()end
	}
}

game.state = game.states.none

local message = require "game.message"
game.message = message

local inputDisplay = require "game.input_display"
game.inputDisplay = inputDisplay

local transit, transitioned, transitSpeed
local nextState, stateArg
local queue = {}

function game.setState(id, ...)
	log("Transitioning to state:", id)

	if transit then
		if transit > 1 then
			transit = 2-transit
		else
			transit = 0
		end
	else
		transit = 0
	end
	transitioned = false
	transitSpeed = 6
	nextState = game.states[id] or id

	if ... == game.noInit then
		stateArg = nil
		log("\tSkipping state initialization!")
	else
		stateArg = {...}
	end
end

function game.setStateInstantly(id, ...)
	game.setState(id, ...)
	log("\tSkipping first half of transition...")
	transit = 1-1e-99
end

function game.queueTask(task, ...)
	queue[#queue+1] = {task, ...}
end

function game.setTransitSpeed(n)
	transitSpeed = n
end

local speed = 1
function game.setSpeed(n)
	speed = n
end
function game.getSpeed()
	return speed
end

function game.update(dt)
	if transit == nil then
		game.state.update(dt * speed)
	else
		if transit == 1 and not transitioned then
			game.setSpeed(1)
			transitioned = true

			for i,task in ipairs(queue) do
				local s,r = pcall(unpack(task))
				if not s then
					log("\tError executing task:", r)
				end
				queue[i] = nil
			end

			if game.state ~= nextState and game.state.deinit then
				game.state.deinit()
			end
			game.state = nextState

			if stateArg then
				nextState.init(unpack(stateArg))
			end

			collectgarbage()
			love.timer.step()
		end

		transit = transit + dt * transitSpeed

		if transit > 1 and not transitioned then
			transit = 1
		elseif transit > 1 and transit < 2 then
			game.state.update(dt * speed)
		elseif transit >= 2 then
			transit = nil
		end
	end

	message.update(dt)
end

function game.draw(dt)
	game.state.draw(dt)

	message.draw()

	if transit ~= nil then
		-- Transitioning animation
		love.graphics.setColor(colors.black(1, 1, 1, 1 - abs(1 - transit)))
		love.graphics.rectangle("fill", 0, 0, window.w, window.h)
		love.graphics.setColor(colors.white)
	end

	if inputDisplay.active then
		inputDisplay.draw(window.w*0.5-39*window.s, 0, window.s*3)
	end

	if demoV then
		love.graphics.setColor(colors.black)
		love.graphics.printf(demoV, 1, window.h-9, window.w, "right")
		love.graphics.setColor(colors.white)
		love.graphics.printf(demoV, 0, window.h-10, window.w, "right")
	end
end

return game
