local love = love
local floor, ceil, round, roundn, max, min, abs, sin, pi = math.floor, math.ceil, math.round, math.roundn, math.max, math.min, math.abs, math.sin, math.pi
local assert, pcall, require, type, pairs, ipairs = assert, pcall, require, type, pairs, ipairs
local window, input, sounds, colors = window, input, sounds, colors
local random = random

local log = newLog "[JNR]"

local jnr = {}

jnr.reset  = "const: jnr.reset"
jnr.reload = "const: jnr.reload"

-- Loading all primary Object Types and Classes
local Tiles = require "jnr.tiles"
local Particle = require "jnr.particle"
local Object = require "jnr.object"
local Entity = require "jnr.entity"
local Player = require "jnr.player"
local Finish = require "jnr.finish"
local WarpDoor = require "jnr.warp_door"
local CheckPoint = require "jnr.check_point"
local SceneTrigger = require "jnr.scene_trigger"
local Barrier = require "jnr.barrier"
local Platform = require "jnr.platform"
local Switch = require "jnr.switch"

jnr.Tiles = Tiles
jnr.Particle = Particle
jnr.Object = Object
jnr.Entity = Entity
jnr.Player = Player
jnr.Finish = Finish
jnr.WarpDoor = WarpDoor
jnr.CheckPoint = CheckPoint
jnr.SceneTrigger = SceneTrigger
jnr.Barrier = Barrier
jnr.Platform = Platform
jnr.Switch = Switch

-- Entity Data Getter
jnr.getEntityData = assets("entityData", function(id)
	local path = "jnr/entity/"..tostring(id)..".lua"
	assert(love.filesystem.isFile(path), "Cannot get entityData of id '"..tostring(id).."'")
	local n = love.filesystem.load(path)()
	n.id = id
	return n
end)
-- Player Data Getter
jnr.getPlayerData = assets("playerData", function(id)
	local path = "jnr/char/"..tostring(id)..".lua"
	local n
	if love.filesystem.isFile(path) then
		n = love.filesystem.load(path)()
	else
		path = "jnr/char/"..tostring(id)..".dpp2c"
		assert(love.filesystem.isFile(path), "Cannot get playerData of id '"..tostring(id).."'")
		local s
		s, n = pcall(Player.parseData, love.filesystem.read(path), false)
		assert(s, n)
	end
	n.id = id
	return n
end)

jnr.swapSkin = false

local tiles, player, player2, finish, particle
local scroll, currentLevel, nextLevel
local hud, runningScene, collectText, levelName
local scrollOverride

jnr.clearConditions = {
	onlyPlayer = function() return #tiles.object == 1 end,
	playerIsDead = function()
		if (player and player.dead) or (player2 and player2.dead) then
			save.data.deaths = save.data.deaths - 1
			return true
		else
			return false
		end
	end
}
jnr.clearCondition = nil

function jnr.resetAll()
	player, player2, finish = nil
	scroll, currentLevel, nextLevel = nil
	runningScene, collectText, levelName = nil
	scrollOverride = nil

	jnr.clearCondition = nil
	jnr.player, jnr.player2, jnr.scroll, jnr.inHub, jnr.currentLevel = nil

	if particle then particle:clear() end
end

local shake = {
	power = 0,
	x = 0, y = 0, r = 0
}
function jnr.shake(n)
	shake.power = min(shake.power+n, 2.5)
end

function jnr.scrollOverride(x, y, speed)
	if x and y then
		if not scrollOverride then
			scrollOverride = {
				x = x,
				y = y,
				speed = speed or 1.5
			}
		else
			scrollOverride.x = x
			scrollOverride.y = y
			scrollOverride.speed = speed or 1.5
		end
	else
		scrollOverride = nil
	end
end

function jnr.getBossBar()
	-- Getting a boss-health-bar-setting-function
	if hud then
		hud.icon.bossfg:setViewport(10, 42, 115, 6)
		local last = 1
		local function bossBar(fill)
			if fill ~= last then
				hud.icon.bossfg:setViewport(10, 42, 115*fill, 6)

				if fill < last then
					hud.bossShake = min(hud.bossShake + 30*(last-fill), 2)
				end
				last = fill
			end
		end

		hud.boss = true
		hud.bossShake = 0

		return bossBar
	else
		return function() end
	end
end

-- Canvas, for rendering shadows
local canvas = {
	init = false,
	active = false,
	update = function(self)
		if self.init then
			if cnf.shadow > 0 then
				self.x_step = cnf.shadow * window.s/window.w
				self.y_step = cnf.shadow * window.s/window.h
				self.bgdist = 6 / cnf.shadow
				self.fgdist = 18/ cnf.shadow

				self.shader:send("x_step", self.x_step)
				self.shader:send("y_step", self.y_step)

				self.active = true
				return
			end
		end
		self.active = false
	end
}
jnr.canvas = canvas

--[[
Particle System
Generating the particle system and particle types
]]
particle = Particle(assets.image("assets/particles.png"), cnf.pmax)
:newType("dust", {{1,1,4,4}, {5,1,4,3}, {9,1,3,3}, {12,1,2,2}}, -60, 60, -10, 0, 0.5, 0.7,
-- Dust. I hate dust.
function(self, p, dt)
	p.xv = p.xv * 0.04 ^ dt
	p.yv = -20 + (20 + p.yv) * 0.7 ^ dt
	p.s = 1 - p.a / p.ma
end, function(self, p, xv, yv, ov)
	p.r = 2*pi * self.random:random() - pi
	p.rv = 6*p.r

	if ov then
		local oy = p.yv
		p.yv = 0.35*p.xv
		p.xv = ov*abs(oy)
		p.ma = p.ma * 0.7
	end

	if xv then p.xv = p.xv + xv end
	if yv then p.yv = p.yv + yv end
end)
:newType("smoke", {{16,1,4,4}, {20,1,4,3}, {24,1,3,3}, {27,1,2,2}}, -90, 90, -20, 20, 2.0, 2.8,
-- Smoking Hot.
function(self, p, dt)
	p.xv = p.xv * 0.08 ^ dt
	p.yv = -20 + (20 + p.yv) * 0.7 ^ dt
	p.s = (1 - p.a / p.ma) * 3
end, function(self, p, xv, yv)
	p.r = 2*pi * self.random:random() - pi
	p.rv = 3*p.r

	if xv then p.xv = p.xv + xv end
	if yv then p.yv = p.yv + yv end
end)
:newType("blood", {{1,7,4,4}, {5,7,4,3}, {9,7,3,3}, {12,7,2,2}, {14,7,2,2}, {16,7,3,2}, {19,7,3,3}, {22,7,4,3}}, -80, 80, -160, 0, 30, 30,
-- Blood
function(self, p, dt)
	if p.st then
		p.s = 1 - p.a / p.ma
	else
		p.yv = 512 - (512 - p.yv) * 0.5 ^ dt
		local coll = self.tiles:getTile(ceil(p.x/self.tiles.ts), ceil(p.y/self.tiles.ts)).coll
		if coll == "full" or coll == "harm" then
			p.st = true
			p.a  = 0
			p.ma = self.random:random()*5+5
			p.xv, p.yv, p.rv = 0, 0, 0
		elseif coll == "bouncy" then
			p.xv = -p.xv
			p.yv = -p.yv * 1.5
			p.rv = -p.rv
		end
	end
end, function(self, p, xv, yv)
	p.r = 2*pi * self.random:random() - pi
	p.rv = 12*p.r
	p.st = false
	p.dt = 0

	if xv then p.xv = p.xv + xv end
	if yv then p.yv = p.yv + yv end
end)
:newType("water", {{31,1,2,2}, {31,1,1,2}, {31,1,1,1}, {32,1,1,2}, {32,2,1,1}}, -60, 60, -60, 0, 1.2, 1.6,
-- Water
function(self, p, dt)
	p.xv = p.xv * 0.8 ^ dt
	p.yv = 512 - (512 - p.yv) * 0.55 ^ dt
	p.s = 1 - p.a / p.ma
end, function(self, p, xv, yv)
	p.r = 2*pi * self.random:random() - pi
	p.rv = 6*p.r

	if xv then p.xv = p.xv + xv end
	if yv then p.yv = p.yv + yv end
end)
:newType("confetti", {{1,12,4,2}, {6,12,4,2}, {11,12,4,2}, {16,12,4,2}, {21,12,4,2}, {26,12,4,2}}, -30, 30, -180, -90, 30, 30,
-- For when life is great!
function(self, p, dt)
	if p.fr then
		p.x0 = p.x0 + p.nv * dt

		if p.yv > 0 then
			p.x = p.x0 + p.xm * sin(p.a * p.sp)
			p.yv = 80 - (80 - p.yv) * 0.95 ^ dt
			p.xv = 0

			local coll = self.tiles:getTile(ceil(p.x/self.tiles.ts), ceil(p.y/self.tiles.ts)).coll
			if coll == "full" or coll == "plat" then
				p.fr = false

				p.a  = 0
				p.ma = self.random:random()*15+10
				p.xv, p.yv, p.rv = 0, 0, 0
			end
		else
			p.yv = 80 - (80 - p.yv) * 0.05 ^ dt
			if p.yv > 0 then
				p.x0 = p.x
				p.xm = p.xv
				p.nv = self.random:random()*4-2
				p.sp = self.random:random()*2+1

				p.xv = 0
				p.a = 0
			end
		end
	else
		p.s = 1 - p.a / p.ma
	end
end, function(self, p)
	p.x0 = p.x
	p.xm = p.xv
	p.rv = self.random:random()*4-2
	p.nv = 0
	p.sp = 1

	p.fr = true
end)
jnr.particle = particle

-- INIT
function jnr.init(levelId, justFin)
	scrollOverride = nil
	if levelId == jnr.reset then
		-- Resetting only
		tiles:resetNow()

		scroll = {
			x = (player.x + player.w*0.5), y = (player.y + player.h*0.5),
			toX = 0, toY = 0
		}
		shake.power = 0
		jnr.scroll = scroll
	elseif levelId == jnr.reload then
		-- Reloading the current level
		jnr.init(currentLevel)
	else
		runningScene = false
		-- Loading and Initialization
		log("Initializing...")

		particle:clear()
		if particle.max ~= cnf.pmax then
			particle:setMax(cnf.pmax)
		end

		if not hud then
			log("\tLoading HUD... (1 time)")
			hud = {
				boss = false
			}

			hud.image = assets.image("assets/jnr_hud.png")
			hud.batch = love.graphics.newSpriteBatch(hud.image, 6, "stream")
			hud.text  = love.graphics.newText(font)

			hud.icon = {
				deaths  = love.graphics.newQuad( 0,  0,  13, 13, 128, 128),
				collect = love.graphics.newQuad(13,  0,  13, 13, 128, 128),
				timer   = love.graphics.newQuad( 0, 13,  13, 13, 128, 128),
				uhc     = love.graphics.newQuad(13, 13,  13, 13, 128, 128),
				bossbg  = love.graphics.newQuad( 0, 27, 128, 12, 128, 128),
				bossfg  = love.graphics.newQuad(10, 42, 115,  6, 128, 128), -- Render: x+10, y+3
			}
		elseif hud.boss then
			hud.boss = false
		end

		if not canvas.init and cnf.shadow > 0 then
			log("\tCreating shader canvases... (1 time)")
			local s, r = pcall(assets.shader, "assets/shader/shadows.frag")
			if s then
				canvas.fg = love.graphics.newCanvas(window.w, window.h)
				canvas.bg = love.graphics.newCanvas(window.w, window.h)
				canvas.shader = r

				canvas.init = true
				canvas:update()
			else
				log("", "", r)
				game.message.queue{colors.red, "Could not load shadow shader.\nPerhaps your graphics card does not support it?"}
			end
		elseif canvas.init and cnf.shadow <= 0 then
			log("\tDeleting shader canvases...")
			canvas.bg = nil
			canvas.fg = nil
			canvas.shader = nil

			canvas.init = false
			canvas:update()
		end

		--[[
		Loading the level etc...
		]]
		log("\tLoading level...")
		local leveldata = love.filesystem.load("jnr/level/"..levelId..".lua")()
		leveldata.id = levelId

		if save.data.hardcore and love.filesystem.isFile("jnr/level/"..levelId.."#uhc.lua") then
			local addleveldata = love.filesystem.load("jnr/level/"..levelId.."#uhc.lua")()

			for k,v in pairs(addleveldata) do
				leveldata[k] = v
			end
			leveldata.id = levelId.."#uhc"
		end

		local isHub = leveldata.hub or false

		-- Player position, saving...
		if isHub then
			if not jnr.inHub and save.data.hub == levelId and #save.data.coord >= 2 then
				leveldata.spawn = save.data.coord
			elseif jnr.inHub then
				save.data.coord = leveldata.spawn
			end
			save.data.hub = levelId

			save.write()
		-- WarpDoor also sets save.data.coord if in a hub.
		end

		nextLevel = leveldata.nextLevel or save.data.hub
		currentLevel = levelId
		if isHub then
			levelName = fileT.require("jnr/level/#names.txt")[levelId] or levelId
		else
			levelName = {
				save.data.flags[levelId] and colors.pGreen or colors.pRed,
				fileT.require("jnr/level/#names.txt")[levelId] or levelId
			}
		end

		-- Creating Level and Tiles
		tiles = Tiles(leveldata)

		if tiles.music ~= nil then
			music.set(tiles.music)
		end

		tiles.particle = particle
		particle.tiles = tiles

		-- Loading the player
		local char = leveldata.char or cnf.char
		player = Player(leveldata.spawn[1], leveldata.spawn[2], 0, jnr.getPlayerData(char))
		if jnr.swapSkin and player.__data.imageP2 then player.anim.img = assets.image(player.__data.imageP2) end
		if leveldata.spawn[3] then player.dir = leveldata.spawn[3] end
		tiles:addObject(player)

		if cnf.multi then
			local char2 = char == "denniz" and "diana" or char == "diana" and "denniz" or char
			player2 = Player(leveldata.spawn[1]+1, leveldata.spawn[2], 0, jnr.getPlayerData(char2))
			if not jnr.swapSkin and player2.__data.imageP2 then player2.anim.img = assets.image(player2.__data.imageP2) end
			if leveldata.spawn[3] then player2.dir = leveldata.spawn[3] end
			player2.updateInputs = player2.updateInputs2
			tiles:addObject(player2)
		else
			player2 = nil
		end

		if not player2 then
			scroll = {
				x = (player.x + player.w*0.5),
				y = (player.y + player.h*0.5),
				toX = 0, toY = 0
			}
		else
			scroll = {
				x = (player.x + player.w*0.5 + player2.x + player2.w*0.5)*0.5,
				y = (player.y + player.h*0.5 + player2.y + player2.h*0.5)*0.5,
				toX = 0, toY = 0
			}
		end

		if justFin then
			if save.data.hardcore then
				particle:spawn("blood"   , 25, player.x + player.w*0.5, player.y, player.w, player.h)
				particle:spawn("confetti", 55, player.x + player.w*0.5, player.y, player.w, player.h)
			else
				particle:spawn("confetti", 80, player.x + player.w*0.5, player.y, player.w, player.h)
			end
		end

		if leveldata.finish then
			finish = Finish(leveldata.finish[1], leveldata.finish[2], -2)

			tiles:addObject(finish)
		else
			finish = nil
		end

		if leveldata.object then
			for k,v in pairs(leveldata.object) do
				if (v.c == nil or save.data.flags[v.c]) and (v.cn == nil or not save.data.flags[v.cn]) then
					if v[1] == "Entity" then
						tiles:addObject(Entity(v[2], v[3], -1, jnr.getEntityData(v[4]), v[5], v[6], v[7], v[8], v[9]))
					else
						tiles:addObject((jnr[v[1]] or v[1])(v[2], v[3], -1, v[4], v[5], v[6], v[7], v[8], v[9]))
					end
				end
			end
		end

		if leveldata.clearIf then
			jnr.clearCondition = jnr.clearConditions[leveldata.clearIf] or leveldata.clearIf
		end

		tiles.canvas = canvas

		tiles:backup()

		local sceneId, sceneArg
		if leveldata.scene then
			if type(leveldata.scene) == "table" then
				for i,v in ipairs(leveldata.scene) do
					if (v.c == nil or save.data.flags[v.c]) and (v.cn == nil or not save.data.flags[v.cn]) then
						sceneId, sceneArg = v[1], v.arg
						if v.f then save.data.flags[v.f] = true end
						break
					end
				end
			else
				sceneId = leveldata.scene
			end
		end

		do
			-- Text displayed in Hubs for the total amount of collected stuff.
			local collect = save.data.totalCollect
			local totalCollect = jnr.collect.__total or 0/0
			collectText = {
				colors.white, ("x%d/%d (%d%%)"):format(collect, totalCollect, 100*collect/totalCollect),
				colors.yellow, " [Total]"
			}
		end

		shake.power = 0
		jnr.player, jnr.player2, jnr.tiles, jnr.scroll, jnr.hud, jnr.inHub, jnr.currentLevel = player, player2, tiles, scroll, hud, isHub, currentLevel
		-- Jnr Init over...

		-- Triggering a scene if necessary.
		if sceneId then
			local inJnr = false
			if type(sceneArg) ~= "table" then
				log("\t\tScene args not a table. Try again later.")
				sceneArg = nil
			elseif sceneArg.j then
				inJnr = true
			end

			if type(sceneId) == "string" then
				log("\tTriggering scene...")
				if inJnr then
					jnr.triggerScene(sceneId, sceneArg)
				else
					game.setStateInstantly("scene", sceneId, sceneArg)
				end
			else
				log("\tScene data is not a string! What the heck?")
			end
		end
	end
end

function jnr.triggerScene(new, arg)
	log("Starting scene in JNR!")
	scene.init(new, arg)
	runningScene = true
end

function jnr.update(dt)
	if input.new("pause") then
		game.setState("menu", "pause", "pause")
	elseif input.new("cselect") and not tiles.charFix then
		game.setState("menu", "char", "pause")
	else
		tiles:update(dt)

		local scrollSpeed = 1.5
		if scrollOverride then
			scroll.toX = min(max(scrollOverride.x, 0), tiles.width*tiles.ts)
			scroll.toY = min(max(scrollOverride.y, 0), tiles.height*tiles.ts)
			if scrollOverride.speed then scrollSpeed = scrollOverride.speed end
		elseif not player2 then
			scroll.toX = min(max(player.x + player.w*0.5 + (player.x-player.pdata.x)/(dt*window.sM), 0), tiles.width*tiles.ts)
			scroll.toY = min(max(player.y + player.h*0.5 + (player.y-player.pdata.y)/(dt*2*window.sM), 0), tiles.height*tiles.ts)
		else
			scroll.toX = min(max((player.x + player.w*0.5 + player2.x + player2.w*0.5 + (player.x-player.pdata.x+player2.x-player2.pdata.x)/(dt*window.sM))*0.5, 0), tiles.width*tiles.ts)
			scroll.toY = min(max((player.y + player.h*0.5 + player2.y + player2.h*0.5 + (player.y-player.pdata.y+player2.y-player2.pdata.y)/(dt*2*window.sM))*0.5, 0), tiles.height*tiles.ts)
		end
		scroll.x = scroll.x + scrollSpeed * window.sM * dt * (scroll.toX - scroll.x)
		scroll.y = scroll.y + scrollSpeed * window.sM * dt * (scroll.toY - scroll.y)

		sounds.setListener(scroll.x, scroll.y)

		if tiles.queueReset then
			tiles.queueReset = false
			game.setState("jnr", jnr.reset)
		end

		save.data.time = save.data.time + dt

		if shake.power > 0 then
			shake.x = (random:random()*2-1)*shake.power
			shake.y = (random:random()*4-2)*shake.power
			shake.r = (random:random()*0.1-0.05)*shake.power

			shake.power = shake.power - dt
			if shake.power < 0 then shake.power = 0 end
		end

		if hud.boss then
			if hud.bossShake > 0 then
				hud.bossShake = hud.bossShake - dt
				if hud.bossShake < 0 then hud.bossShake = 0 end
			end
		end

		if runningScene then
			if scene.progressScript(dt, false) then
				runningScene = false
			end
		end

		if jnr.clearCondition then
			-- Checking level clear condition
			local cleared
			if type(jnr.clearCondition) == "function" then
				cleared = jnr.clearCondition()
			else
				cleared = true
			end

			if cleared then
				-- Level is cleared, do stuff
				jnr.clearCondition = nil

				local alreadyCleared = save.data.flags[currentLevel]
				local priorCollect, thisCollect

				if tiles.maxCollect then
					priorCollect = save.data.collect[currentLevel] or 0
					thisCollect  = tiles.collect or 0
					if thisCollect > priorCollect then
						save.data.collect[currentLevel] = thisCollect
						save.data.totalCollect = save.data.totalCollect + thisCollect - priorCollect
					end
				end

				save.data.flags[currentLevel] = true

				if thisCollect and thisCollect > priorCollect then
					game.queueTask(game.message.queue, {
						colors.softBlue, "Cleared: ",
						colors.white, fileT.require("jnr/level/#names.txt")[currentLevel] or currentLevel,
						colors.white, "\nGot ",
						colors.yellow, tostring(thisCollect-priorCollect),
						colors.white, " new Collectibles!",
						thisCollect >= tiles.maxCollect and colors.yellow or nil, thisCollect >= tiles.maxCollect and "\n100% Collected!" or nil
					})
				else
					game.queueTask(game.message.queue, {
						colors.softBlue, "Cleared: ",
						colors.white, fileT.require("jnr/level/#names.txt")[currentLevel] or currentLevel,
					})
				end

				-- Blah, blah, next level, blah... etc.
				game.setState("jnr", nextLevel, true)
			end
		end
	end
end

function jnr.draw()
	love.graphics.push()

	love.graphics.scale(window.s)

	local lx = roundn(-scroll.x + window.vw*0.5, window.s)
	local ly = roundn(-scroll.y + window.vh*0.5, window.s)
	tiles:drawBackground(lx, ly, window.vw, window.vh)

	if shake.power > 0 then
		local s = (shake.power*0.05+1) * window.s
		love.graphics.scale(shake.power*0.05+1)
		love.graphics.translate(roundn(window.w/s*0.5, s), roundn(window.h/s*0.5, s))
		love.graphics.rotate(shake.r)
		love.graphics.translate(roundn(-scroll.x + shake.x, s), roundn(-scroll.y + shake.y, s))
	else
		love.graphics.translate(lx, ly)
	end

	tiles:draw(-lx, -ly, window.vw, window.vh, window.s)

	if game.state == jnr then
		if runningScene then
			love.graphics.translate(-lx, -ly)
			scene.drawBox(lx, ly)
			love.graphics.translate(lx, ly)
		end

		hud.batch:clear()
		hud.text :clear()

		hud.text :add(levelName, 3, 3)

		hud.batch:add(hud.icon.collect, 1, 16)
		if tiles.maxCollect then
			hud.text:add(("x%d/%d"):format(tiles.collect, tiles.maxCollect), 15, 18)
		else
			hud.text:add(collectText, 15, 18)
		end

		hud.batch:add(save.data.hardcore and hud.icon.uhc or hud.icon.deaths, 1, 31)
		hud.text :add(("x%d"):format(save.data.deaths), 15, 33)

		hud.batch:add(hud.icon.timer, 1, 46)
		hud.text :add(save.formatTime(save.data.time), 15, 48)

		if hud.boss then
			local r = 0
			if hud.bossShake > 0 then
				r = (random:random()*0.2-0.1)*hud.bossShake
			end
			hud.batch:add(hud.icon.bossbg, window.vw*0.5, window.vh-12, r, 2, 2, 64, 6)
			hud.batch:add(hud.icon.bossfg, window.vw*0.5, window.vh-12, r, 2, 2, 54, 3)
		end

		love.graphics.setColor(colors.black)
		love.graphics.draw(hud.text , 1-lx, 1-ly)
		love.graphics.draw(hud.batch, 1-lx, 1-ly)
		love.graphics.setColor(colors.white)
		love.graphics.draw(hud.batch, -lx, -ly)
		love.graphics.draw(hud.text , -lx, -ly)
	end

	love.graphics.pop()
end

function jnr.resize(w, h)
	if canvas.init then
		canvas.fg = love.graphics.newCanvas(window.w, window.h)
		canvas.bg = love.graphics.newCanvas(window.w, window.h)
	end

	canvas:update()
end

-- Some general area data
jnr.area = {
	[1] = {
		hub = "1-hub",
		name = "Nearby Forest",
		door = 2
	},
	[2] = {
		hub = "2-hub",
		name = "Cinder City",
		door = 3
	},
	[3] = {
		hub = "3-hub",
		name = "Invasion Island",
		door = 4
	},
	[4] = {
		hub = "4-hub",
		name = "Rainbow Factory",
		door = 5
	},
	[5] = {
		hub = "5-hub",
		name = "The Final Gambit",
		door = 6
	},
	nan_city = {
		hub = "nan_city",
		name = "Needs-a-Name City",
		door = 3
	},
	the_end = {
		hub = "the_end",
		name = "???",
		door = 1
	}
}
if _demo then
	jnr.area[_demo] = jnr.area.the_end
end

-- Collectable amounts
jnr.collect = fileT.load("jnr/level/#collect.txt")

do
	local exclude = {
		["old_1-1"] = true,
		["old_1-2"] = true,
		["old_1-3"] = true,
		["1-5"] = true
	}

	local totalCollect = 0
	for k,v in pairs(jnr.collect) do
		if not exclude[k] then
			totalCollect = totalCollect + v
		end
	end

	jnr.collect.__total = totalCollect
end

return jnr
