local floor, ceil, abs, roundn, huge = math.floor, math.ceil, math.abs, math.roundn, math.huge
local remove = table.remove
local type, pairs, ipairs, getmetatable = type, pairs, ipairs, getmetatable
local sfx = sfx

local Object = {}

function Object:new(x, y, w, h, z)
	self.x, self.y = x, y
	self.w, self.h = w, h
	self.xv,self.yv= 0, 0
	self.z = z or 0
	self.dir = 1

	self.mass = self.w * self.h

	self.data = {
		ground = false,
		ceiling= false,
		wall   = false,
		moving = false,
		water  = false,
		duck   = false,
		onTop  = false
	}

	self.pdata = {
		ground = false,
		ceiling= false,
		wall   = false,
		moving = false,
		water  = false,
		duck   = false,
		x  = x, y  = y,
		xv = 0, yv = 0
	}

	self.xvLimit = 0
	self.xvAcc   = 1.5e-7
	self.yvLimit = 512
	self.yvAcc   = 0.4

	self.dustTime = 0
	self.canColide = true
	self.canCollect = false
	self.floating = false

	self.jumped = false
	self.__overlapping = false
end Object.__new = Object.new

function Object:init()
	-- Called when added to a Tiles Object.
end

local function copyTable(t)
	local n = {}
	for k,v in pairs(t) do
		if type(k) ~= "string" or not k:find("^__") then
			if type(v) ~= "table" or getmetatable(v) then
				n[k] = v
			else
				n[k] = copyTable(v)
			end
		end
	end
	return n
end

function Object:backup()
	self.__backup = copyTable(self)
end

function Object:reset()
	if self.__backup then
		for k,v in pairs(self.__backup) do
			if type(v) ~= "table" or getmetatable(v) then
				self[k] = v
			else
				self[k] = copyTable(v)
			end
		end
	end
end Object.__reset = Object.reset

function Object:getStatus()
	return (self.data.ground and 1 or 0) + (self.data.wall and 2 or 0) + (self.data.water and 4 or 0) + (self.data.duck and 8 or 0)
end

function Object:setAcceleration()
	-- Acceleration (Gravity/Air/Ground)
	if self.data.ground then
		self.xvAcc = 1.5e-7
	else
		self.xvAcc = 0.008
	end

	if self.data.wall then
		self.yvLimit = 30
		self.yvAcc   = 6e-7
	else
		self.yvLimit = 512
		self.yvAcc   = 0.4
	end
end

function Object:accelerate(dt)
	if self.data.water then
		local xv = self.xvLimit * 0.5
		local yv = self.yvLimit * 0.5

		self.xv = xv - (xv - self.xv) * (self.xvAcc * 0.5) ^ (dt * 0.5)
		self.yv = yv - (yv - self.yv) * (self.yvAcc * 0.5) ^ (dt * 0.5)
	else
		local xv = self.xvLimit
		local yv = self.yvLimit

		self.xv = xv - (xv - self.xv) * self.xvAcc ^ dt
		self.yv = yv - (yv - self.yv) * self.yvAcc ^ dt

		if abs(self.xv) < dt * 5 then
			self.xv = 0
		end
	end
end

local function slopeCode(self, dt, y, cx, cy, ts, xc, yc, cc, uy, m)
	if y < uy then
		if self.tiles:getTile(cx+xc, cy+yc).coll ~= cc then
			y = uy
		end
	end

	local sub = 0
	if (m > 0 and self.xv < 0) or (m < 0 and self.xv > 0) then
		sub = m * self.xv * dt * 2
	end

	if self.y + self.yv * dt + self.h - sub >= y then
		self.y = y - self.h
		self.data.ground = true

		if self.yv > 0 then
			self.yv = 0
		end
	end
end

function Object:colide(dt)
	local data  = self.data
	local pdata = self.pdata

	-- Update PData (Past Data) and other stuff
	if data.water and not pdata.water then
		self.xv = self.xv * 0.5
		self.yv = self.yv * 0.5
	end

	pdata.ground = data.ground
	pdata.ceiling= data.ceiling
	pdata.wall   = data.wall
	pdata.moving = data.moving
	pdata.water  = data.water
	pdata.duck   = data.duck
	pdata.x  = self.x
	pdata.y  = self.y
	pdata.xv = self.xv
	pdata.yv = self.yv

	-- Collision
	local tiles = self.tiles
	local level = tiles.level
	local ts = tiles.ts

	local object = tiles.object

	-- Prevent collision calculation with already overlapped objects
	for k, other in ipairs(object) do
		if other ~= self then
			if self:overlap(other) then
				other.__overlapping = true
				self:hasColided(other, "n", dt)
				other:hasColided(self, "n", dt)
			else
				other.__overlapping = false
			end
		end
	end

	-- Resetting these
	data.ground = false
	data.ceiling= false
	data.wall   = false

	--[[
	X Collision
	]]
	local xv = self.xv
	if data.onTop then
		local onTop = data.onTop
		data.onTop = false

		local i = 0
		while onTop and i < 32 do
			xv = xv + onTop.xv
			onTop = onTop.data.onTop
			i = i + 1
		end
	end
	if xv ~= 0 then
		-- Limit velocity
		if xv > 5000 then
			xv = 5000
		elseif xv < -5000 then
			xv = -5000
		end

		local minx, maxx, step, closest
		if xv > 0 then
			minx = ceil((self.x + self.w) / ts)
			maxx = ceil((self.x + xv * dt + self.w) / ts)
			step = 1
			closest = huge
		else
			maxx = ceil((self.x + xv * dt) / ts)
			minx = ceil((self.x) / ts)
			step = -1
			closest = -huge
		end
		local miny = ceil((self.y + 0.1) / ts)
		local maxy = ceil((self.y + self.h - 0.1) / ts)

		-- Tile collision X
		local cx, cy, tile

		for x=minx, maxx, step do
			local xL = level[x]
			if xL then for y=miny, maxy do
				local t = xL[y]
				if t then
					if self.canCollect and t.collect then
						tiles.collect = tiles.collect + 1
						tiles:setTile(x, y, t.collect)
						sfx.collect:play(x*ts-ts*0.5, y*ts-ts*0.5)
					elseif t.coll ~= "none" and t.coll ~= "plat" and ((xv > 0 and x < closest) or (xv < 0 and x > closest) or (x == closest and t.prio > tile.prio)) then
						closest = x
						tile = t
						cx, cy = x, y
					end
				end
			end end
		end

		if tile then
			local coll = tile.coll
			if coll == "full" then
				if xv > 0 then
					self.x = cx*ts-ts - self.w
					self.dir = 1
				else
					self.x = cx*ts
					self.dir = -1
				end

				self.xv = 0
				xv = 0
				data.wall = true
			elseif coll == "harm" then
				if xv > 0 then
					self.x = cx*ts-ts - self.w
				else
					self.x = cx*ts
				end

				xv = -xv
				self.xv = xv
				self:hurt(tile)
			elseif coll == "bouncy" then
				if xv > 0 then
					self.x = cx*ts-ts - self.w
				else
					self.x = cx*ts
				end

				xv = -xv * 2
				self.xv = xv
				data.wall = 0
			elseif coll == "custom" then
				local nxv, nyv = (tile.colideX or tile.colide)(self, tiles, cx, cy, dt, xv, self.yv)
				if nxv ~= nil then xv = nxv self.xv = nxv end
				if nyv ~= nil then self.yv = nyv end
			end
		end

		self.x = self.x + xv * dt

		-- Object collision X
		for k, other in ipairs(object) do
			if other ~= self then
				if not other.__overlapping and self:overlap(other) then
					if self.canColide and other.canColide then
						local a, b
						if self.mass > other.mass then
							b, a = self, other
						else
							a, b = self, other
						end
						local nxv = (a.xv * a.mass + b.xv * b.mass)/(a.mass + b.mass)
						a.xv = nxv
						b.xv = nxv

						if a.x - (self == a and xv * dt or 0) < b.x then
							if nxv < 0 and a.data.wall then
								b.x = a.x + a.w
							else
								a.x = b.x - a.w
							end

							if a.data.ground or b.data.ground then
								a.data.wall = true
								a.dir = 1
								b.data.wall = true
								b.dir = -1
							end

							a:hasColided(b, "x+", dt)
							b:hasColided(a, "x-", dt)
						else
							if nxv > 0 and a.data.wall then
								b.x = a.x - b.w
							else
								a.x = b.x + b.w
							end

							if a.data.ground or b.data.ground then
								a.data.wall = true
								a.dir = -1
								b.data.wall = true
								b.dir = 1
							end

							a:hasColided(b, "x-", dt)
							b:hasColided(a, "x+", dt)
						end
					else
						other:hasColided(self, "x", dt)
						self:hasColided(other, "x", dt)
					end
				end
			end
		end
	end
	--[[
	X Collision End
	Y Collision Begin
	]]--
	if self.yv ~= 0 then
		if data.water then data.water = false end

		-- Limit velocity
		if self.yv > 5000 then
			self.yv = 5000
		elseif self.yv < -5000 then
			self.yv = -5000
		end

		local minx = ceil((self.x + 0.1) / ts)
		local maxx = ceil((self.x + self.w - 0.1) / ts)
		local miny, maxy, closest
		if self.yv > 0 then
			miny = ceil((self.y + self.h) / ts)
			maxy = ceil((self.y + self.yv * dt + self.h) / ts)
			closest = huge
		else
			miny = ceil((self.y + self.yv * dt) / ts)
			maxy = ceil((self.y) / ts)
			closest = -huge
		end

		-- Tile collision Y
		local cx, cy, tile

		for x=minx, maxx do
			local xL = level[x]
			if xL then for y=miny, maxy do
				local t = xL[y]
				if t then
					if self.canCollect and t.collect then
						tiles.collect = tiles.collect + 1
						tiles:setTile(x, y, t.collect)
						sfx.collect:play(x*ts-ts*0.5, y*ts-ts*0.5)
					elseif t.water and not data.water then
						data.water = true
					elseif t.coll ~= "none" and not (t.coll == "plat" and (self.yv < 0 or self.y + self.h > y*ts-ts)) and ((self.yv > 0 and y < closest) or (self.yv < 0 and y > closest) or (y == closest and t.prio > tile.prio)) then
						closest = y
						tile = t
						cx, cy = x, y
					end
				end
			end end
		end

		if tile then
			local coll = tile.coll
			if coll == "full" then
				if self.yv > 0 then
					self.y = cy*ts-ts - self.h
					data.ground = true
				else
					self.y = cy*ts
					data.ceiling = true
				end

				self.yv = 0

				if data.water and not pdata.water then data.water = false end
			elseif coll == "plat" then
				local y = cy*ts-ts
				if self.y + self.h <= y and not (self.data.duck and self.jumped) then
					self.y = y - self.h
					data.ground = true

					self.yv = 0

					if data.water and not pdata.water then data.water = false end
				end
			elseif coll == "harm" then
				if self.yv > 0 then
					self.y = cy*ts-ts - self.h
				else
					self.y = cy*ts
				end

				self.yv = -self.yv
				self:hurt(tile)
			elseif coll == "bouncy" then
				if self.yv > 0 then
					self.y = cy*ts-ts - self.h + self.yv * 1.5 * dt
					data.ground = 0
				else
					self.y = cy*ts + self.yv * 1.5 * dt
					data.ceiling = 0
				end

				self.yv = -self.yv * 1.5
			elseif coll == "slopel" then
				local y = cy*ts-ts + (self.x - (cx*ts-ts))
				slopeCode(self, dt, y, cx, cy, ts, -1, -1, "slopel", cy*ts-ts, -1)
			elseif coll == "sloper" then
				local y = cy*ts - (self.x + self.w - (cx*ts-ts))
				slopeCode(self, dt, y, cx, cy, ts, 1, -1, "sloper", cy*ts-ts, 1)
			elseif coll == "slopell" then
				local y = cy*ts-ts*0.5 + (self.x - (cx*ts-ts))*0.5
				slopeCode(self, dt, y, cx, cy, ts, -1, 0, "slopelu", cy*ts-ts*0.5, -.5)
			elseif coll == "slopelu" then
				local y = cy*ts-ts + (self.x - (cx*ts-ts))*0.5
				slopeCode(self, dt, y, cx, cy, ts, -1, -1, "slopell", cy*ts-ts, -.5)
			elseif coll == "sloperl" then
				local y = cy*ts - (self.x + self.w - (cx*ts-ts))*0.5
				slopeCode(self, dt, y, cx, cy, ts, 1, 0, "sloperu", cy*ts-ts*0.5, 0.5)
			elseif coll == "sloperu" then
				local y = cy*ts-ts*0.5 - (self.x + self.w - (cx*ts-ts))*0.5
				slopeCode(self, dt, y, cx, cy, ts, 1, -1, "sloperl", cy*ts-ts, 0.5)
			elseif coll == "custom" then
				local nxv, nyv = (tile.colideY or tile.colide)(self, tiles, cx, cy, dt, self.xv, self.yv)
				if nxv ~= nil then self.xv = nxv end
				if nyv ~= nil then self.yv = nyv end
			end
		end

		self.y = self.y + self.yv * dt

		-- Object collision Y
		for k, other in ipairs(object) do
			if other ~= self then
				if not other.__overlapping and self:overlap(other) then
					if self.canColide and other.canColide then
						local a, b
						if self.mass > other.mass then
							b, a = self, other
						else
							a, b = self, other
						end
						local nyv = (a.yv * a.mass + b.yv * b.mass)/(a.mass + b.mass)
						a.yv = nyv
						b.yv = nyv

						if a.y < b.y then
							if nyv < 0 and a.data.ceiling then
								b.y = a.y + a.h
							else
								a.y = b.y - a.h
							end

							if not a.floating then a.data.onTop = b end
							a.data.ground = true
							b.data.ceiling = true

							b:hasColided(a, "y-", dt)
							a:hasColided(b, "y+", dt)
						else
							if nyv > 0 and a.data.ground then
								b.y = a.y - b.h
								a.yv = 0
								b.yv = 0
							else
								a.y = b.y + b.h
							end

							if not b.floating then b.data.onTop = a end
							b.data.ground = true
							a.data.ceiling = true

							b:hasColided(a, "y+", dt)
							a:hasColided(b, "y-", dt)
						end
					else
						other:hasColided(self, "y", dt)
						self:hasColided(other, "y", dt)
					end
				end
			end
		end
	end
end

function Object:hurt(who)
	-- Code goes here...
end

function Object:hasColided(other, dir, dt)
	-- Code goes here...
end

function Object:overlap(other)
	return self.x + self.w > other.x and
		self.y + self.h > other.y and
		other.x + other.w > self.x and
		other.y + other.h > self.y
end

function Object:remove()
	for i=1, #self.tiles.object do
		if self.tiles.object[i] == self then
			remove(self.tiles.object, i)
		end
	end
end

function Object:spawnParticles(dt)
	local data  = self.data
	local pdata = self.pdata

	if data.ground then
		if not pdata.ground then
			-- Hitting the ground
			self.tiles.particle:spawn("dust", pdata.yv*self.mass*0.0003, self.x + self.w * 0.5, self.y + self.h, self.w * 0.3, 1)
		elseif not self.floating and (abs(self.xv) > 45 or data.moving) then
			-- RUN dust
			self.dustTime = self.dustTime + abs(self.xv) * dt * 0.25
			if self.dustTime > 1 then
				self.tiles.particle:spawn("dust", self.dustTime, self.x + self.w * 0.5, self.y + self.h, self.w * 0.3, 1)

				self.dustTime = self.dustTime % 1
			end
		end
	elseif data.ceiling then
		if not pdata.ceiling then
			-- Hitting the ceiling
			self.tiles.particle:spawn("dust", -pdata.yv*self.mass*0.0003, self.x + self.w * 0.5, self.y, self.w * 0.3, 1, 0, 10)
		end
	end

	if data.wall then
		if not pdata.wall then
			-- Hitting a wall
			if pdata.xv > 0 then
				self.tiles.particle:spawn("dust", pdata.xv*self.mass*0.0006, self.x + self.w, self.y + self.h * 0.5, 1, self.h * 0.3, 0, 0, -1)
			else
				self.tiles.particle:spawn("dust", -pdata.xv*self.mass*0.0006, self.x, self.y + self.h * 0.5, 1, self.h * 0.3, 0, 0, 1)
			end
		elseif abs(self.yv) > 40 then
			-- Running/Sliding on a wall
			self.dustTime = self.dustTime + abs(self.yv) * dt * 0.5
			if self.dustTime > 1 then
				if pdata.xv > 0 then
					self.tiles.particle:spawn("dust", self.dustTime, self.x + self.w, self.y + self.h * 0.5, 1, self.h * 0.3, 0, 0, -1)
				else
					self.tiles.particle:spawn("dust", self.dustTime, self.x, self.y + self.h * 0.5, 1, self.h * 0.3, 0, 0, 1)
				end

				self.dustTime = self.dustTime % 1
			end
		end
	end

	if data.water ~= pdata.water then
		if self.yv > 0 then
			self.tiles.particle:spawn("water", self.yv*self.mass*0.0015, self.x + self.w * 0.5, roundn(self.y + self.h, 1/self.tiles.ts) + 3, self.w * 0.3, 1, -self.xv*0.5, -self.yv*0.5)
		else
			self.tiles.particle:spawn("water", -self.yv*self.mass*0.0015, self.x + self.w * 0.5, roundn(self.y, 1/self.tiles.ts) + 3, self.w * 0.3, 1, self.xv*0.5, self.yv*0.5)
		end
	end
end

function Object:makeSounds()
	local data  = self.data
	local pdata = self.pdata

	if data.wall and not pdata.wall then
		local vol = self.mass * abs(pdata.xv) * 1.0e-5
		sfx.colide:play(self.x + (pdata.xv > 0 and self.w or 0), self.y + self.h*0.5, vol)
	end

	if (data.ground and not pdata.ground) or (data.ceiling and not pdata.ceiling) then
		local vol = self.mass * abs(pdata.yv) * 1.0e-5
		sfx.colide:play(self.x + self.w*0.5, self.y + (pdata.yv > 0 and self.h or 0), vol)
	end
end

Object.kill = Object.remove

function Object:update(dt)
	self:setAcceleration()
	self:accelerate(dt)
	self:colide(dt)
	self:spawnParticles(dt)
	self:makeSounds()
end

function Object:draw()
	love.graphics.rectangle("fill", self.x, self.y, self.w, self.h)
end

function Object:__tostring()
	return ("%s: Pos. %.2f, %.2f"):format(self.__class.type or "Object", self.x or 0/0, self.y or 0/0)
end

function Object:__inherited(cls)
	cls.base.type = cls.name
end

Object.type = "Object"

return class("Object", Object)
