local love = love
local remove = table.remove
local min = math.min

local Particle = {}

function Particle:new(image, max)
	self.max = max
	self.image = image
	self.batch = love.graphics.newSpriteBatch(image, max, "stream")
	self.imgw, self.imgh = image:getDimensions()

	self.type = {}
	self.part = {}

	self.random = assets.random("particle")
end

function Particle:setMax(max)
	self.max = max
	self.batch = love.graphics.newSpriteBatch(self.image, max, "stream")
end

function Particle:clear()
	self.part = {}
end

function Particle:newType(id, quadBase, xvm, xv, yvm, yv, mag, ag, update, init)
	local quads = {}
	for k,v in pairs(quadBase) do
		local ox = v[3] * 0.5
		local oy = v[4] * 0.5
		quads[#quads+1] = {
			quad = love.graphics.newQuad(v[1], v[2], v[3], v[4], self.imgw, self.imgh),
			ox = ox, oy = oy
		}
	end

	local data = {
		id = id, quads = quads,
		av = mag, ava = ag-mag,
		xv = xvm, xva = xv-xvm,
		yv = yvm, yva = yv-yvm,
		update = update, init = init
	}
	self.type[id] = data

	return self
end

function Particle:spawn(id, n, x, y, xo, yo, ...)
	local data = self.type[id]
	if data then
		local l = #self.part
		for i=l+1, min(l+n, self.max) do
			local q = data.quads[self.random:random(1, #data.quads)]

			local p = {
				id = id,
				update = data.update, quad = q.quad,
				x  = x + self.random:random()*xo*2-xo,
				y  = y + self.random:random()*yo*2-yo,
				xv = data.xv + self.random:random()*data.xva,
				yv = data.yv + self.random:random()*data.yva,
				ma = data.av + self.random:random()*data.ava,
				ox = q.ox, oy = q.oy,
				r = 0, rv = 0, s = 1, a = 0
			}
			if data.init then data.init(self, p, ...) end

			self.part[i] = p
		end
	end
end

function Particle:spawnSingle(id, x, y, ...)
	local data = self.type[id]
	if data then
		local l = #self.part
		if l >= self.max then return end

		local q = data.quads[self.random:random(1, #data.quads)]

		local p = {
			id = id,
			update = data.update, quad = q.quad,
			x = x, y = y,
			xv = data.xv + self.random:random()*data.xva,
			yv = data.yv + self.random:random()*data.yva,
			ma = data.av + self.random:random()*data.ava,
			ox = q.ox, oy = q.oy,
			r = 0, rv = 0, s = 1, a = 0
		}
		if data.init then data.init(self, p, ...) end

		self.part[l+1] = p

		return p
	end
end

function Particle:update(dt)
	for i=#self.part, 1, -1 do
		local p = self.part[i]
		p.a = p.a + dt
		if p.a > p.ma then
			remove(self.part, i)
		else
			p.update(self, p, dt)
			p.x = p.x + p.xv * dt
			p.y = p.y + p.yv * dt
			p.r = p.r + p.rv * dt
		end
	end
end

function Particle:draw(dt, x, y)
	local batch = self.batch

	batch:clear()
	for i=1, #self.part do
		local p = self.part[i]
		batch:add(p.quad, p.x, p.y, p.r, p.s, p.s, p.ox, p.oy)
	end
	love.graphics.draw(batch, x, y)
end

function Particle:__tostring()
	return ("Particle: Max %d; Current %d"):format(self.max, #self.part)
end

return class("Particle", Particle)
