local floor, ceil, abs, max, min, roundn = math.floor, math.ceil, math.abs, math.max, math.min, math.roundn
local ipairs = ipairs
local window, input = window, input
local sfx = sfx

local Player = {}

Player.parseData = require "jnr.player_parsedata"

Player.attackAnim = anim.newHeader(assets.image("assets/attack_swipe.png"), 13, 19)
	:reg(0, 1, 1, 9, 60, false)

function Player:new(x, y, z, data)
	data = data or {}

	local w = data.w or 8
	local h = data.h or 17

	self:__new(x, y, w, h, z)

	self.dir = 1
	self.jumps = 0
	self.pJumps = data.pJumps or 1
	self.countGJump = false if data.countGJump ~= nil then self.countGJump = data.countGJump end

	self.jump = {
		normal = data.normalJump or -210,
		wall   = data.wallJump   or -180,
		air    = data.airJump    or -180,
		water  = data.swimVert   or -120
	}
	self.speed = {
		normal = data.walkSpeed or 120,
		crawl  = data.crawlSpeed or 45,
		wall   = data.wallSpeed or 180,
		water  = data.swimHori  or 140
	}

	self.canSwim = true if data.canSwim ~= nil then self.canSwim = data.canSwim end
	self.canWallRun = true if data.canWallRun ~= nil then self.canWallRun = data.canWallRun end
	self.canWallJump = true if data.canWallJump ~= nil then self.canWallJump = data.canWallJump end

	self.ctrl = {}

	self.code = data.code or {}

	self.dead   = false
	self.frozen = false
	self.jumped = false

	self.statusFix = false
	self.dustTime = 0
	self.canCollect = true

	self.attacking = false
	self.attackDir = 1
	self.attackAnim = self.attackAnim:clone():set(0)
	self.attackHitbox = false

	if data.anim then
		self.anim = data.anim:clone()
	else
		self.anim = anim.newHeader(assets.image("assets/char/denniz.png"), 11, 19)
	end

	self.xrOff = data.xrOff or floor(self.w * 0.5)
	self.xrCen = data.xrCen or floor(self.anim.tx * 0.5)
	self.yrOff = data.yrOff or 1

	self.playerId = data.id

	self.__data = data
	if data.new then
		data.new(self)
	end
end

function Player:init()
	if self.__data.init then
		self.__data.init(self)
	end
end

function Player:hurt()
	if self.dead then
		self.tiles.particle:spawn("blood", 10, self.x + self.w*0.5, self.y + self.h*0.5, self.w*0.5, self.h*0.5, self.xv, self.yv)
	else
		self.tiles.particle:spawn("blood", 40, self.x + self.w*0.5, self.y + self.h*0.5, self.w*0.5, self.h*0.5, self.xv, self.yv)
		self:kill()
	end
end

function Player:kill()
	sfx.death:play()

	self.dead   = 1
	self.frozen = true

	jnr.scrollOverride(
		self.x + self.w*0.5,
		self.y + self.h*0.5,
		2.5
	)
	save.data.deaths = save.data.deaths + 1
end

function Player:reset()
	self:__reset()
	self:animate(0)
end

function Player:duck()
	self.data.duck = true
	local h = self.h * 0.5
	self.y = self.y + h
	self.h = h
end

function Player:unduck()
	self.data.duck = false
	self.y = self.y - self.h
	self.h = self.h * 2
end

function Player:updateInputs()
	local up, down, left, right, holdJump, jump, attack
	if not self.frozen then
		up       = input.down "up"
		down     = input.down "down"
		left     = input.down "left"
		right    = input.down "right"
		holdJump = input.down "jump"
		jump     = input.new  "jump"
		attack   = input.new  "attack"
	else
		up, down, left, right, holdJump, jump, attack = false, false, false, false, false, false, false
	end
	local ctrl = self.ctrl
	ctrl.up, ctrl.down, ctrl.left, ctrl.right, ctrl.holdJump, ctrl.jump, ctrl.attack = up, down, left, right, holdJump, jump, attack
	ctrl.axisL, ctrl.axisR, ctrl.axisU, ctrl.axisD = input.axis "left", input.axis "right", input.axis "up", input.axis "down"
	return up, down, left, right, holdJump, jump, attack
end

function Player:updateInputs2()
	local up, down, left, right, holdJump, jump, attack
	if not self.frozen then
		up       = input.down "up-gp"
		down     = input.down "down-gp"
		left     = input.down "left-gp"
		right    = input.down "right-gp"
		holdJump = input.down "jump-gp"
		jump     = input.new  "jump-gp"
		attack   = input.new  "attack-gp"
	else
		up, down, left, right, holdJump, jump, attack = false, false, false, false, false, false, false
	end
	local ctrl = self.ctrl
	ctrl.up, ctrl.down, ctrl.left, ctrl.right, ctrl.holdJump, ctrl.jump, ctrl.attack = up, down, left, right, holdJump, jump, attack
	ctrl.axisL, ctrl.axisR, ctrl.axisU, ctrl.axisD = input.axis "left-gp", input.axis "right-gp", input.axis "up-gp", input.axis "down-gp"
	return up, down, left, right, holdJump, jump, attack
end

function Player:update(dt)
	-- Getting inputs
	local up, down, left, right, holdJump, jump, attack = self:updateInputs()

	local data  = self.data
	local pdata = self.pdata

	-- Whether or not the player is ducking... and changing the hitbox accordingly.
	local needToDuck = data.ground and down
	if not needToDuck and data.ground then
		local topCoord = 1+floor((self.y-(data.duck and self.h or 0))/self.tiles.ts)
		for x=1+floor((self.x)/self.tiles.ts), ceil((self.x+self.w)/self.tiles.ts) do
			local coll = self.tiles:getTile(x, topCoord).coll
			if coll == "full" or coll == "harm" then
				needToDuck = true
				break
			end
		end
	end
	if data.duck then
		if not needToDuck then
			self:unduck()
		end
	elseif needToDuck then
		self:duck()
	end

	if data.water and self.canSwim then
		--[[
		WATER
		]]
		-- Acceleration (Gravity/Air/Ground)
		if data.ground then
			self.xvAcc = 1.5e-7
		else
			self.xvAcc = 0.008
		end

		self.yvLimit = 512
		self.yvAcc   = 0.4

		-- Player Input (Walking/Jumping)
		if right ~= left then
			local speed = data.duck and self.speed.crawl or data.ground and self.speed.normal or self.speed.water
			if right then
				local axis = self.ctrl.axisR
				if axis ~= 0 then speed = speed * axis end

				self.xvLimit = speed
				self.dir = 1
			else
				local axis = self.ctrl.axisL
				if axis ~= 0 then speed = speed * axis end

				self.xvLimit = -speed
				self.dir = -1
			end
			data.moving = true
		else
			self.xvLimit = 0
			data.moving = false
		end
	else
		--[[
		LAND/AIR
		]]
		-- Acceleration (Gravity/Air/Ground)
		if data.ground then
			if data.duck and abs(self.xv) > self.speed.crawl then
				self.xvAcc = 0.4
			else
				self.xvAcc = 1.5e-7
			end
		else
			self.xvAcc = 0.008
		end

		if data.wall then
			if up and self.canWallRun and self.yv < 0 then
				self.yvLimit = 23
				self.yvAcc   = 0.16
			elseif down or self.yv < 0 then
				self.yvLimit = 170
				self.yvAcc   = 0.07
			else
				self.yvLimit = 30
				self.yvAcc   = 6e-7
			end
		else
			self.yvLimit = 512
			self.yvAcc   = 0.4
		end

		-- Player Input (Walking/Jumping)
		if not (data.duck and abs(self.xv) > self.speed.crawl) and (right ~= left) then
			local speed = data.duck and self.speed.crawl or self.speed.normal
			if right then
				local axis = self.ctrl.axisR
				if axis ~= 0 then speed = speed * axis end

				self.xvLimit = speed
				if data.ground then self.dir = 1 end
			else
				local axis = self.ctrl.axisL
				if axis ~= 0 then speed = speed * axis end

				self.xvLimit = -speed
				if data.ground then self.dir = -1 end
			end
			data.moving = true
		else
			self.xvLimit = 0
			data.moving = false
		end
	end

	if self.code[self.anim.anim] then self.code[self.anim.anim](self, dt) end
	if self.code.always then self.code.always(self, dt) end

	self:accelerate(dt)

	if self.jumped then
		if self.canSwim and pdata.water and not data.water then
			self.yv = self.yv - 50
		elseif data.ground or data.wall or self.yv > 0 or not holdJump then
			self.jumped = false
			if not holdJump then
				self.yv = self.yv * 0.45
			end
		end
	end

	if data.water and self.canSwim then
		--[[
		WATER
		]]
		self.jumps = 0

		if jump then
			if down then
				self.yv = self.jump.water * 0.5
			else
				self.yv = self.jump.water
			end

			if self.anim.anims.swim2 then
				self.statusFix = self:getStatus()
				self.anim:set("swim2"):reset()
			end

			self.jumped = true
		end
	else
		--[[
		LAND/AIR
		]]
		if data.ground then
			self.jumps = 0
		elseif data.wall then
			self.jumps = 0
		end

		if jump then
			self.jumped = true
			if data.duck then
			elseif data.ground then
				-- Jumping from the ground (normal)
				self.tiles.particle:spawn("dust", -self.jump.normal*0.02, self.x + self.w * 0.5, self.y + self.h, self.w * 0.3, 1, -self.xv * 0.25)
				sfx.colide:play(self.x + self.w*0.5, self.y + self.h, 0.075)

				if self.countGJump then self.jumps = self.jumps + 1 end
				self.yv = self.jump.normal
			elseif self.canWallJump and data.wall then
				-- Kicking off a wall (wall)
				if self.dir > 0 then
					self.tiles.particle:spawn("dust", -self.jump.wall*0.02, self.x + self.w, self.y + self.h * 0.5, 1, self.h * 0.3, 0, 0, -1)
					sfx.colide:play(self.x + self.w, self.y + self.h*0.5, 0.075)

					self.xv = -self.speed.wall
				else
					self.tiles.particle:spawn("dust", -self.jump.wall*0.02, self.x, self.y + self.h * 0.5, 1, self.h * 0.3, 0, 0, 1)
					sfx.colide:play(self.x, self.y + self.h*0.5, 0.075)

					self.xv = self.speed.wall
				end
				if self.countGJump then self.jumps = self.jumps + 1 end
				self.yv = self.jump.wall
				self.dir = -self.dir
			elseif self.jumps < self.pJumps then
				-- Jumping in mid-air (air)
				self.tiles.particle:spawn("dust", max((self.yv-self.jump.air)*0.02, 3), self.x + self.w * 0.5, self.y + self.h, self.w * 0.3, 1, -self.xv * 0.25)

				self.jumps = self.jumps + 1
				self.yv = self.jump.air

				if self.xvLimit > 0 then
					self.dir = 1
				elseif self.xvLimit < 0 then
					self.dir = -1
				end

				if self.anim.anims.jumpd then
					self.statusFix = self:getStatus()
					self.anim:set("jumpd"):reset()
				end
			end
		end

		if up and self.canWallRun and data.wall and not pdata.wall then
			local toyv
			if data.ground then
				toyv = self.yv*0.25 - abs(pdata.xv)
			else
				toyv = self.yv*0.25 - abs(pdata.xv)*0.75
			end
			if toyv < self.yv then
				self.yv = toyv
			end
		end
	end

	-- Collision in object.lua
	self:colide(dt)

	if self.attacking then
		if self.attackAnim:update(dt) then
			self.attacking = false
			self.attackHitbox = false
		else
			if self.dir ~= self.attackDir then
				self.dir = self.attackDir
			end

			if (self.xvLimit < 0 and self.dir > 0) or (self.xvLimit > 0 and self.dir < 0) then
				self.xvLimit = self.xvLimit * 0.75
			end

			local hitbox = self.attackHitbox
			hitbox.x = self.x + (self.dir > 0 and self.w or -13)
			hitbox.y = self.y - (data.duck and self.h or 0)

			for k, other in ipairs(self.tiles.object) do
				if other.type ~= "Player" and not hitbox.hit[other] then
					if other:overlap(hitbox) then
						other.xv = (45000 * self.attackDir + other.mass * other.xv)/(100 + other.mass)
						other.yv = (10000 + other.mass * other.yv)/(100 + other.mass)
						other:hurt(self)

						hitbox.hit[other] = true
					end
				end
			end
		end
	elseif attack and not data.duck then
		self.attackDir = self.dir

		if self.anim.anims.attack and not data.duck then
			self.statusFix = self:getStatus()
			self.anim:set("attack"):reset()
		end

		sfx.swing:play(self.x + (self.dir > 0 and self.w or 0), self.y + self.h*0.5)

		local hitbox = {
			x = self.x + (self.dir > 0 and self.w or -13),
			y = self.y - (data.duck and self.h or 0),
			w = 13,
			h = (data.duck and self.h*2 or self.h),
			hit = {}
		}
		self.attackHitbox = hitbox

		for k, other in ipairs(self.tiles.object) do
			if other.type ~= "Player" then
				if other:overlap(hitbox) then
					other.xv = (45000 * self.attackDir + other.mass * other.xv)/(100 + other.mass)
					other.yv = (10000 + other.mass * other.yv)/(100 + other.mass)
					other:hurt(self)

					hitbox.hit[other] = true
				end
			end
		end

		self.attacking = 1
		self.attackAnim:reset():update(dt)
	end

	self:animate(dt)

	self:dying(dt)

	-- Particles & Sounds
	self:spawnParticles(dt)
	self:makeSounds()
end

function Player:dying(dt)
	-- Being dead. Period.
	if self.dead then
		self.dead = self.dead - dt / game.getSpeed()
		if self.dead > 0 then
			game.setSpeed(min(0.2 + self.dead*0.8, 1))
		else
			self.tiles:reset()
		end
	elseif self.y > self.tiles.height * self.tiles.ts + window.vh * 0.5 then
		self:kill()
	end
end

function Player:animate(dt)
	local data = self.data
	local anim = self.anim
	-- Animation
	if self.dead then
		if data.ground and anim.anims.deadg then
			anim:set "deadg"
		else
			anim:set "dead"
		end

		anim:update(dt)
	elseif self.statusFix then
		if anim:update(dt) or self.statusFix ~= self:getStatus() then
			self.statusFix = false
		end
	else
		local anims = anim.anims

		if data.duck then
			if data.moving and anims.crawl then
				anim:set "crawl"
			elseif abs(self.xv) > self.speed.crawl and anims.slide then
				anim:set "slide"
			else
				anim:set "duck"
			end
		elseif data.ground then
			if data.moving then
				if data.wall and anims.push then
					anim:set "push"
				else
					anim:set "walk"
				end
			elseif anims.idle_long and (anim.anim == "idle_long" or anim.loops > 42 and anim.anim == "idle") then
				anim:set "idle_long"
			else
				anim:set "idle"
			end
		elseif data.water and anims.swim then
			anim:set "swim"
		elseif data.wall then
			if self.canWallRun and self.yv < 0 then
				anim:set "wallrun"
			else
				anim:set "wall"
			end
		else
			if anims.jumpf and ((self.dir > 0 and self.xv > 45) or (self.dir < 0 and self.xv < -45)) then
				anim:set "jumpf"
			elseif anims.jumpb and ((self.dir > 0 and self.xv < -45) or (self.dir < 0 and self.xv > 45)) then
				anim:set "jumpb"
			else
				anim:set "jump"
			end
		end

		if anim.anim == "walk" then
			anim:update(dt * abs(self.xvLimit)/self.speed.normal)
		elseif anim.anim == "crawl" then
			anim:update(dt * abs(self.xvLimit)/self.speed.crawl)
		elseif anim.anim == "wallrun" then
			anim:update(dt * -self.yv/45)
		else
			anim:update(dt)
		end
	end
end

function Player:draw(s)
	self.anim:draw(roundn(self.x + self.xrOff, s), roundn(self.y - (self.data.duck and self.h or 0), s), 0, self.dir, 1, self.xrCen, self.yrOff)

	if self.attackHitbox then
		local hitbox = self.attackHitbox
		self.attackAnim:draw(roundn(hitbox.x + (self.attackDir > 0 and 0 or hitbox.w), s), roundn(hitbox.y, s), 0, self.attackDir, hitbox.h/17, 0, 1)
	end
end

return class("Player", require("jnr.object"), Player)
