local love = love
local floor, ceil, max, min, abs = math.floor, math.ceil, math.max, math.min, math.abs
local sort = table.sort
local assert, type, tostring, tonumber = assert, type, tostring, tonumber
local unpack, setmetatable, rawset = unpack, setmetatable, rawset
local pairs, ipairs, pcall = pairs, ipairs, pcall
local colors = colors

--[[
Tiles aka Level class
]]
local Tiles = {}
local getTileData

function Tiles:new(arg)
	if arg then
		assert(type(arg.tiles) == "string", "Invalid tile set path.")
		self.__tileData = getTileData(arg.tiles)
		for k,v in pairs(self.__tileData) do
			self[k] = v
		end

		self:setLevel(arg)

		self.queueReset = false

		self.charFix = arg.char and true or false
		if arg.music ~= nil then self.music = arg.music end
	end
end

getTileData = assets("tileData", function(id)
	local path = "jnr/tiles/"..id..".lua"
	assert(love.filesystem.isFile(path), "Cannot find tile set file: "..path)

	local raw = love.filesystem.load(path)()

	local new = Tiles()
	new:setTiles(raw)
	new.tilesId = id

	return new
end)

function Tiles:setTiles(arg)
	self.ts = tonumber(arg.ts) or 9

	if arg.image == nil then arg.image = "forest" end
	assert(type(arg.image) == "string", "No valid image path.")

	local fullImagePath = "assets/tiles/"..arg.image..".png"
	assert(love.filesystem.isFile(fullImagePath), "Cannot find image file: "..fullImagePath)
	self.image = assets.image(fullImagePath)

	assert(arg.data, "No Tile Data found!")
	self.tiles = {}
	self.animated = {}

	self.header = anim.newMultiHeader(self.image, self.ts, self.ts)
	self.batch = love.graphics.newSpriteBatch(self.image, 7000, "stream")

	-- Loading/Parsing Tile Data
	for k,v in pairs(arg.data) do
		local id = #self.tiles + 1
		self.tiles[id] = self:newTileData(id, unpack(v))
	end

	-- Loading Background
	self.bg = {}
	if arg.bg then self:setBackground(arg.bg) end

	-- Loading constant shader for all drawing operations
	if arg.shader then
		local s, r = pcall(self.setShader, self, arg.shader)
		if not s then print(r) end
	end

	if arg.music ~= nil then
		self.music = arg.music
	end
end

function Tiles:setShader(arg)
	local shader = setmetatable({
		shader = assets.shader("assets/shader/"..tostring(arg.path)..".frag"),
		time  = false,
		scale = false,
	}, {})
	assert(shader.shader ~= nil, "Could not load tile shader.")

	for k,v in pairs(arg) do
		if k ~= "path" then
			if k == "time" or k == "scale" or k == "x" or k == "y" then
				shader[k] = v
			elseif type(v) == "table" then
				if #v == 4 then
					shader.shader:sendColor(k, v)
				else
					shader.shader:send(k, v)
				end
			else
				shader.shader:send(k, v)
			end
		end
	end

	self.shader = shader
end

Tiles.tileMeta = {
	__tostring = function(self)
		return ("TileData: Id %d; Var. %d"):format(self.id or 0/0, self.varId or 0)
	end,
	type = "tile"
}
Tiles.tileMeta.__index = Tiles.tileMeta

Tiles.defaultCollPriority = {
	full = 10, plat = 9,
	harm = 8, bouncy = 8,
	sloper = 6, slopel = 6,
	slopelu = 7, slopell = 5,
	sloperu = 7, sloperl = 5,
	none = 0, custom = 1
}
Tiles.presetColl = assets("tileColl", function(id)
	assert(type(id) == "string", "presetColl ID needs to be a string")
	local path = "jnr/tiles/preset/"..id..".lua"
	assert(love.filesystem.isFile(path), "Cannot find presetColl of ID '"..id.."'")
	return love.filesystem.load(path)()
end)

function Tiles:newTileData(id, x, y, arg, var)
	local data = setmetatable({}, self.tileMeta)

	data.id = id

	for k,v in pairs(arg) do
		if type(k) == "number" then
			if v == "collect" then
				data.collect = 0
			else
				data[v] = true
			end
		elseif k == "anim" then
			-- TX, TY, Lenght, FPSSpeed, DoLoop
			data.anim = self.header:reg(unpack(v))
		elseif k == "anim2" then
			data.anim = self.header:spc(unpack(v))
		else
			data[k] = v
		end
	end

	data.quad = self.header:getQuad(x, y)
	if data.coll == nil then data.coll = "none" end
	if data.bg   == nil then data.bg   = false  end
	if data.prio == nil then data.prio = self.defaultCollPriority[data.coll] or 0 end

	if var then
		data.var = {}
		for k,v in pairs(var) do
			local ndata = self:newTileData(id, unpack(v))
			data.var[#data.var+1] = ndata
			ndata.varId = #data.var
		end
	end

	if data.anim then
		self.animated[#self.animated+1] = data
	end

	return data
end

function Tiles:setBackground(bg)
	assert(type(bg) == "table", "Invalid background data.")

	self.bg = {}
	for i,v in ipairs(bg) do
		local file, scroll, dirs = v[1], v[2], v[3]
		assert(type(file) == "string", "Invalid background image path in layer "..tostring(i)..".")
		assert(type(scroll) == "number", "Invalid background scroll factor in layer "..tostring(i)..".")
		assert(type(dirs) == "string", "Invalid background scroll direction string in layer "..tostring(i)..".")

		local fullImagePath = "assets/background/"..file..".png"
		assert(love.filesystem.isFile(fullImagePath), "Cannot find image file: "..fullImagePath)

		local autoScroll, offset

		local img = assets.image(fullImagePath)
		local w, h = img:getDimensions()
		local xs, ys, scale
		if dirs:find("x") then xs = true else xs = false end
		if dirs:find("y") then ys = true else ys = false end
		if xs and ys then
			scale = v[4]
			assert(type(scale) == "number", "Invalid background XY scaling factor in layer "..tostring(i)..".")

			if v[5] then
				autoScroll = v[5] or {0, 0}
				assert(type(autoScroll) == "table", "Invalid background autoscroll data in layer "..tostring(i)..".")
				assert(type(autoScroll[1]) == "number" and type(autoScroll[2]) == "number", "Invalid background autoscroll data in layer "..tostring(i)..".")
			end
			offset = {0, 0}
		else
			if v[4] then
				autoScroll = v[4]
				assert(type(autoScroll) == "number", "Invalid background autoscroll in layer "..tostring(i)..".")
			end
			offset = 0
		end

		self.bg[#self.bg+1] = {
			batch = love.graphics.newSpriteBatch(img, xs and ys and 6400 or 80, "stream"),
			img = img, w = w, h = h,
			xs = xs, ys = ys, scale = scale, auto = autoScroll,
			scroll = scroll, offset = offset
		}
	end
end

function Tiles:setLevel(arg)
	self.levelId = arg.id
	-- Assinging a level to the tiles
	self.level = setmetatable({}, {
		__index = function(t, k)
			local v = {}
			rawset(t, k, v)
			return v
		end
	})

	-- Copying/Converting the level data into Tile-Data
	local width  = 0
	local height = 0
	local level = arg.level

	local collect
	if arg.collect then
		if type(arg.collect) == "number" then
			collect = arg.collect
		else
			collect = 0
		end
	end

	for y=1, #level do
		local yL = level[y]
		if yL then
			if y > height then height = y end
			for x=1, #yL do
				if x > width then width = x end
				local t = self.tiles[ yL[x] ]
				if t then
					if t.var then
						local v = (x + (width+4999) * (y-1)) % (#t.var+1)
						if v == 0 then
							self.level[x][y] = t
						else
							self.level[x][y] = t.var[v]
						end
					else
						self.level[x][y] = t
					end

					if collect then
						if t.collect then
							collect = collect + 1
						end
					end
				else
					self.level[x][y] = false
				end
			end
		end
	end

	self.maxCollect = collect
	self.collect = 0

	self.width = width
	self.height = height

	self.failTile = setmetatable({
		quad = self.header:getQuad(1, 1),
		coll = "none",
		bg = true,
		prio = -1,
		id = -1
	}, self.tileMeta)

	if arg.out then
		-- Optional argument to allow a certain tile to be used outside the level range
		local out = self.tiles[ arg.out ]

		if out then
			local layermeta = {
				__index = function(t, k)
					rawset(t, k, out)
					return out
				end,
				__newindex = function()end
			}
			local outlayer = setmetatable({}, layermeta)

			for k,v in pairs(self.level) do
				setmetatable(v, layermeta)
			end

			setmetatable(self.level, {
				__index = function(t, k)
					rawset(t, k, outlayer)
					return outlayer
				end
			})
		else
			print("Tile ID "..tostring(arg.out).." does not exist and cannot be used as 'out'.")
			setmetatable(self.level, nil)
		end
	elseif arg.extend then
		local layermeta = {
			__index = function(t, k)
				local tile
				if k > self.height then
					tile = rawget(t, self.height)
				else
					tile = rawget(t, 1)
				end
				rawset(t, k, tile)
				return tile
			end
		}

		for k,v in pairs(self.level) do
			setmetatable(v, layermeta)
		end

		setmetatable(self.level, {
			__index = function(t, k)
				local layer
				if k > self.width then
					layer = rawget(t, self.width)
				else
					layer = rawget(t, 1)
				end
				rawset(t, k, layer)
				return layer
			end
		})
	else
		setmetatable(self.level, nil)
	end

	if arg.bg then self:setBackground(arg.bg) end

	self.object = {}
end

function Tiles:backup()
	local backup = {}
	for x,layer in pairs(self.level) do
		backup[x] = {}
		for y,tile in pairs(layer) do
			backup[x][y] = tile
		end
		setmetatable(backup[x], getmetatable(layer))
	end
	setmetatable(backup, getmetatable(self.level))

	self.__level = backup

	self.__object = {}
	for k,v in pairs(self.object) do
		self.__object[#self.__object+1] = v
		v:backup()
	end

	self.__collect = self.collect
end

function Tiles:reset()
	self.queueReset = true
end

function Tiles:resetNow()
	if self.__level then
		local backup = {}
		for x,layer in pairs(self.__level) do
			backup[x] = {}
			for y,tile in pairs(layer) do
				backup[x][y] = tile
			end
			setmetatable(backup[x], getmetatable(layer))
		end
		setmetatable(backup, getmetatable(self.__level))

		self.level = backup
	end

	if self.__object then
		self.object = {}
		for k,v in pairs(self.__object) do
			self.object[k] = v
			v:reset()
		end
	else
		for k,v in pairs(self.object) do
			v:reset()
		end
	end

	self.collect = self.__collect

	self.particle:clear()
end

function Tiles:getTile(x, y)
	if self.level[x] then
		return self.level[x][y] or self.failTile
	else
		return self.failTile
	end
end

function Tiles:getRawTile(x, y)
	if self.level[x] then
		return self.level[x][y] or false
	else
		return false
	end
end

function Tiles:setTile(x, y, id)
	if self.level[x] then
		local t = self.tiles[id]
		if t then
			if t.var then
				local v = (x + (self.width+4999) * (y-1)) % (#t.var+1)
				if v == 0 then
					self.level[x][y] = t
				else
					self.level[x][y] = t.var[v]
				end
			else
				self.level[x][y] = t
			end
		else
			self.level[x][y] = false
		end
	end
end

local function sortObjects(a, b)
	return a.z < b.z
end

function Tiles:addObject(obj)
	self.object[#self.object+1] = obj

	obj.x = self.ts*obj.x - self.ts*0.5 - obj.w*0.5
	obj.y = self.ts*obj.y - obj.h

	obj.tiles = self

	obj:init()

	sort(self.object, sortObjects)

	return obj
end

function Tiles:addObjects(...)
	for k,obj in pairs{...} do
		self.object[#self.object+1] = obj

		obj.x = self.ts*obj.x - self.ts*0.5 - obj.w*0.5
		obj.y = self.ts*obj.y - self.ts*0.5 - obj.h*0.5

		obj.tiles = self

		obj:init()
	end

	sort(self.object, sortObjects)
end

function Tiles:update(dt)
	self.header:update(dt)

	for i=1, #self.animated do
		local v = self.animated[i]
		v.quad = v.anim.quad
	end

	for k,v in pairs(self.object) do
		v:update(dt)
	end

	self.particle:update(dt)
	self:updateBackground(dt)

	if self.shader then
		local shader = self.shader
		if shader.time then
			shader.time = shader.time + dt
			shader.shader:send("TIME", shader.time)
		end
	end
end

--[[
Background
]]
function Tiles:updateBackground(dt)
	for i=1, #self.bg do
		local layer = self.bg[i]

		if layer.auto then
			if layer.xs and layer.ys then
				layer.offset[1] = layer.offset[1] + layer.auto[1] * dt
				layer.offset[2] = layer.offset[2] + layer.auto[2] * dt
			elseif layer.xs or layer.ys then
				layer.offset = layer.offset + layer.auto * dt
			end
		end
	end
end

function Tiles:drawBackground(x, y, w, h)
	for i=1, #self.bg do
		local layer = self.bg[i]

		if layer.xs and layer.ys then
			local s = layer.scale
			local xstep = layer.w * s
			local mx = (x * layer.scroll + layer.offset[1]) % xstep - xstep

			local ystep = layer.h * s
			local my = (y * layer.scroll + layer.offset[2]) % ystep - ystep

			layer.batch:clear()
			for xc=mx, w, xstep do
				for yc=my, h, ystep do
					layer.batch:add(xc, yc, 0, s)
				end
			end
			love.graphics.draw(layer.batch)
		elseif layer.xs then
			local s = h / layer.h
			local xstep = layer.w * s
			local mx = (x * layer.scroll + layer.offset) % xstep - xstep

			layer.batch:clear()
			for xc=mx, w, xstep do
				layer.batch:add(xc, 0, 0, s)
			end
			love.graphics.draw(layer.batch)
		elseif layer.ys then
			local s = w / layer.w
			local ystep = layer.h * s
			local my = (y * layer.scroll + layer.offset) % ystep - ystep

			layer.batch:clear()
			for yc=my, w, ystep do
				layer.batch:add(0, yc, 0, s)
			end
			love.graphics.draw(layer.batch)
		end
	end
end

--[[
Drawing
]]
Tiles.canvas = { active = false, init = false }
function Tiles:draw(x, y, w, h, s)
	-- Top left coordinates and width + height
	local ts = self.ts
	local level = self.level
	local batch = self.batch
	local canvas = self.canvas
	local shader
	if self.shader then
		shader = self.shader.shader
		if self.shader.scale then shader:send("SCALE", s) end
		if self.shader.x then shader:send("X", x) end
		if self.shader.y then shader:send("Y", y) end
	end

	-- Tile Borders
	local xmin = ceil(x / ts)
	local ymin = ceil(y / ts)
	local xmax = ceil((x + w) / ts)
	local ymax = ceil((y + h) / ts)

	if canvas.active then
		love.graphics.setCanvas(canvas.bg)
		love.graphics.clear(0, 0, 0, 0)
	end
	if shader then
		love.graphics.setShader(shader)
	end

	batch:clear()
	for xt=xmin, xmax do
		local xL = level[xt]
		if xL then for yt=ymin, ymax do
			local t = xL[yt]
			if t then
				if t.bg then
					batch:add(t.quad, xt*ts-ts, yt*ts-ts)
				end
			end
		end end
	end
	love.graphics.draw(batch, 0, 0)

	for i=1, #self.object do
		self.object[i]:draw(s)
	end
	self.particle:draw()

	if canvas.active then
		love.graphics.setCanvas(canvas.fg)
		love.graphics.clear(0, 0, 0, 0)
	end

	batch:clear()
	for xt=xmin, xmax do
		local xL = level[xt]
		if xL then for yt=ymin, ymax do
			local t = xL[yt]
			if t then
				if not t.bg then
					batch:add(t.quad, xt*ts-ts, yt*ts-ts)
				end
			end
		end end
	end
	love.graphics.draw(batch, 0, 0)

	if canvas.active then
		love.graphics.setCanvas(nil)

		love.graphics.push()
		love.graphics.origin()
		love.graphics.setBlendMode("alpha", "premultiplied")

		love.graphics.setShader(canvas.shader)

		canvas.shader:send("dist", canvas.y_step*canvas.bgdist)
		canvas.shader:send("darkness", 0.55)
		love.graphics.draw(canvas.bg)

		canvas.shader:send("dist", canvas.y_step*canvas.fgdist)
		canvas.shader:send("darkness", 0.75)
		love.graphics.draw(canvas.fg)

		love.graphics.setBlendMode("alpha", "alphamultiply")
		love.graphics.pop()

		love.graphics.setShader(shader)
	end

	for i=1, #self.object do
		local v = self.object[i]
		if v.drawL then
			v:drawL(s)
		end
	end

	love.graphics.setShader()

	if debugging then
		self:debugDraw(x, y, w, h, s)
	end
end

--[[
Debug Rendering
(Schematicized rendering)
]]
Tiles.debugColors = {
	harm = colors.new(255, 0, 0, 85), bouncy = colors.new(0, 255, 0, 85),
	none = colors.new(0, 0, 0, 85), collect = colors.new(255, 255, 0, 85),
	other = colors.new(0, 0, 255, 85), entity = colors.new(255, 150, 0, 85)
}

Tiles.debugShapes = {
	plat = {0, 0, 1, 0, 1, 0.2, 0, 0.2},
	sloper = {0, 1, 1, 0, 1, 1, 0, 1},
	slopel = {0, 0, 1, 1, 0, 1, 0, 0},
	slopelu = {0, 0, 1, 0.5, 1, 1, 0, 1},
	slopell = {0, 0.5, 1, 1, 0, 1, 0, 0.5},
	sloperu = {0, 0.5, 1, 0, 1, 1, 0, 1},
	sloperl = {0, 1, 1, 0.5, 1, 1, 0, 1},
	none = {0, 0.1, 0.9, 0, 1, 0.9, 0.1, 1},
	other = {0, 0, 1, 0, 1, 1, 0, 1}
}

function Tiles:debugDraw(x, y, w, h, s)
	-- Top left coordinates and width + height
	local ts = self.ts
	local level = self.level

	-- Tile Borders
	local xmin = floor(x / ts)
	local ymin = floor(y / ts)
	local xmax = ceil((x + w) / ts)
	local ymax = ceil((y + h) / ts)

	-- Tiles
	for xt=xmin, xmax do
		local xL = level[xt]
		if xL then for yt=ymin, ymax do
			local t = xL[yt]
			if t then
				love.graphics.setColor(t.collect and self.debugColors.collect or self.debugColors[t.coll] or self.debugColors.other)

				local shape = self.debugShapes[t.coll] or self.debugShapes.other
				local xx, yy = xt*ts-ts, yt*ts-ts
				love.graphics.polygon("fill", xx+shape[1]*ts, yy+shape[2]*ts, xx+shape[3]*ts, yy+shape[4]*ts, xx+shape[5]*ts, yy+shape[6]*ts, xx+shape[7]*ts, yy+shape[8]*ts)
			end
		end end
	end

	-- Entities
	love.graphics.setColor(Tiles.debugColors.entity)
	for k,v in pairs(self.object) do
		love.graphics.rectangle("fill", v.x, v.y, v.w, v.h)

		if v.attackHitbox then
			local v = v.attackHitbox
			love.graphics.rectangle("fill", v.x, v.y, v.w, v.h)
		end
	end
	love.graphics.setColor(colors.white)
end

if _arg.tiles_debug_only then
	-- If the user says so...
	Tiles.draw = Tiles.debugDraw

	for k,v in pairs(Tiles.debugColors) do
		v[4] = 175
	end
end

-- This function stores an image of the current level, including entities
function Tiles:storeImage(saveTo)
	local w, h = self.width * self.ts, self.height * self.ts
	local target = love.graphics.newCanvas(w, h)

	local origSize = self.batch:getBufferSize()
	self.batch:setBufferSize(999999)

	local ow, oh, os = window.w, window.h, window.s
	window.w, window.h, window.s = w, h, 1
	if self.canvas.active then
		self.canvas.fg = love.graphics.newCanvas(window.w, window.h)
		self.canvas.bg = love.graphics.newCanvas(window.w, window.h)
		self.canvas:update()
	end

	local setCanvas = love.graphics.setCanvas
	function love.graphics.setCanvas(...) if ... == nil then setCanvas(target) else setCanvas(...) end end

	setCanvas(target)
	self:drawBackground(0, 0, w, h)
	self:draw(0, 0, w, h, 1)
	setCanvas(nil)
	love.graphics.setCanvas = setCanvas

	self.batch:setBufferSize(origSize)

	window.w, window.h, window.s = ow, oh, os
	if self.canvas.active then
		self.canvas.fg = love.graphics.newCanvas(window.w, window.h)
		self.canvas.bg = love.graphics.newCanvas(window.w, window.h)
		self.canvas:update()
	end

	return target:newImageData():encode("png", saveTo..".png")
end

function Tiles:__tostring()
	if self.levelId then
		return ("Tiles: Level %s; Tiles %s"):format(self.levelId, self.tilesId or "unknown")
	else
		return ("Tiles: Tiles %s"):format(self.tilesId or "unknown")
	end
end

Tiles = class("Tiles", Tiles)
return Tiles
