--[[
Simple library for loading and discarding data via garbage collector, that may not always be in use.
	loadingStorage = assets(name, constructor)
loadingStorage will also be stored as assets[name].
To get a value from the loadingStorage, do:
	myValue = loadingStorage[keyToGet]
	myValue = loadingStorage(keyToGet, ...)
If the resource is not in memory, it will call the constructor and return its output, meanwhile storing it for
latter re-usage.
'nil' as a key is supported, 'nan' is not.
assets.image, assets.shader and assets.random are predefined.
]]

local rawget, rawset = rawget, rawset
local setmetatable, getmetatable = setmetatable, getmetatable
local type, assert = type, assert

local assets = {}
local __nil = setmetatable({}, { __tostring = function() return "nil" end })

local __constructor = setmetatable({}, { __mode = "k" })

local metaTable = {
	__mode = "kv",
	__call = function(t, k, ...)
		if k == nil then k = __nil end
		if rawget(t, k) ~= nil then
			return rawget(t, k)
		else
			local K
			if k ~= __nil then K = k end
			local new = __constructor[t](K, ...)
			rawset(t, k, new)
			return new
		end
	end,
	__index = function(t, k)
		if k == nil then
			if rawget(t, __nil) ~= nil then
				return rawget(t, __nil)
			end
		end
		local new = __constructor[t](k)
		rawset(t, k == nil and __nil or k, new)
		return new
	end,
	__newindex = function(t, k, v)
		if k == nil then k = __nil end
		rawset(t, k, v)
	end,
	__tostring = function(t)
		return "assetLS ("..tostring(__constructor[t])..")"
	end
}

function assets:__call(name, constructor)
	assert(type(constructor) == "function" or getmetatable(constructor) and getmetatable(constructor).__call, "Argument #2 to 'assets' not callable.")
	local new = setmetatable({}, metaTable)
	__constructor[new] = constructor

	rawset(self, name, new)

	return new
end
assets.__constructor = __constructor
assets.__nil = __nil
assets.__index = assets

assets = setmetatable({}, assets)

assets("image", love.graphics.newImage)
assets("shader", love.graphics.newShader)

local newRandomGenerator, random = love.math.newRandomGenerator, love.math.random
assets("random", function(use)
	return newRandomGenerator(random(0, 0xffffffff), random(0, 0xffffffff))
end)

return assets
