--[[
Simple class implementation.
new = class(name, parent, base)
	'base' stores all default values of the new class and its (or its objects') meta-functions.
	'base.new(obj, ...)' will be used as the constructor of all its objects.
	All arguments but 'base' are optional, but must be supplied in this order.
	Inheriting a class will invoke its '__inherited' field, passing first itself, then the new class.
new:extend(key, value)
	Extends a class and all its children by that value. Discouraged, define the class first.
Contained key: name, new, base, parent, children

obj = new(...)
	Creates an object.
Always has access to the keys '__class' and, if its class has a parent, '__parent'.
]]

local type, tostring = type, tostring
local setmetatable, getmetatable = setmetatable, getmetatable
local rawget, rawset = rawget, rawset
local pairs = pairs

local class = setmetatable({}, {
	__call = function(self, a, b, c)
		return self.new(a, b, c)
	end
})

class.meta = {
	__index = function(self, k)
		local v = rawget(self.base, k)
		if v ~= nil then
			return v
		elseif rawget(self, "parent") ~= nil then
			return self.parent[k]
		else
			return rawget(class.meta, k)
		end
	end,
	__call = function(self, ...)
		local obj = setmetatable({}, self.base)
		self.new(obj, ...)
		return obj
	end,
	__tostring = function(self)
		return "class "..self:definition()
	end,
	definition = function(self)
		if self.parent then
			return tostring(self.name).." : "..self.parent:definition()
		else
			return tostring(self.name)
		end
	end,
	extend = function(self, k, v)
		if k == "__index" then
			local bmeta = getmetatable(self.base)
			if bmeta and rawget(bmeta, "__index") == nil then
				rawset(bmeta, "__index", v)
			else
				setmetatable(self.base, { __index = v })
			end
		elseif rawget(self.base, k) == nil then
			rawset(self.base, k, v)
		end
		if self.children then
			for l,cls in pairs(self.children) do
				cls:extend(k, v)
			end
		end
	end
}
class.meta.__newindex = class.meta.extend

class.weakTable = { __mode = "kv" }

function class.noConstructor() end

function class.new(a, b, c)
	local name, parent, rawbase
	if c == nil then
		if     b == nil then rawbase = a
		elseif a == nil then rawbase = {}
		else
			if type(a) == "string" then
				name, rawbase = a, b
			else
				parent, rawbase = a, b
			end
		end
	else
		name, parent, rawbase = a, b, c
	end

	local bmeta = {}
	local base = {}

	if parent and parent.base then
		for k,v in pairs(parent.base) do
			if not (k == "__index" and v == parent.base) then
				base[k] = v
			end
		end
		local pmeta = getmetatable(parent.base)
		if pmeta and pmeta.__index then
			bmeta.__index = pmeta.__index
		end
	end
	for k,v in pairs(rawbase) do
		if k == "__index" then
			bmeta.__index = v
		else
			base[k] = v
		end
	end

	for k,v in pairs(bmeta) do
		setmetatable(base, bmeta)
		break
	end

	base.__index = base

	local cls = setmetatable({
		base = base,
		new = base.new or parent and parent.new or class.noConstructor,
		name = name or "(unnamed)",
		parent = parent,
		children = setmetatable({}, class.weakTable)
	}, class.meta)

	base.__class = cls
	base.__parent = parent

	if parent and parent.children then
		parent.children[#parent.children+1] = cls
	end

	if parent and parent.__inherited then
		parent:__inherited(cls)
	end

	return cls
end

return class
