--[[
Library which wraps the usual audio library, simplifies it (somewhat) and allows for easier 2-dimensional sound management:
newSFX = sounds.newSFX(path, volume_mod, [id])
	newSFX:play(x_position, y_position) [Note: Absolute positions]
	+ some functions for normal sources
Also includes a simple, somewhat memory-effective wrapper for including music:
newMusic = sounds.newMusic(path, volume_mod, [id])
	newMusic:play()
	+ some functions for normal sources
	This library will only have the current music in memory and will make sure to only play one song at a time.
sounds.setListener(listener_x, listener_y)
	Sets the position of the listener (e.g. camera position).
sounds.enableThreadedLoading()
	Calling this function causes any future newMusic:play() to load the source in another thread.
	Subsequently, it may seem like the call is delayed (which it technically is) but no longer blocks the main thread until the source is loaded.
sounds.update(dt)
	Makes sure music fades out and stops. Usually takes 1 second, though this can be slowed down/sped up by multiplying dt. (2 * dt -> 0.5 seconds fade time)
	If sounds.enableThreadedLoading() has been called before, this will also make sure that the music is activated upon loading.
sounds.setPositionScale(scale)
	Divisor for any position given as an argument to a function.
]]
local function log(...)
	print("[Sounds Main]",...)
end

local pcall, setmetatable, pairs, pcall, type, tonumber = pcall, setmetatable, pairs, pcall, type, tonumber
local loveAudioSetPosition = love.audio.setPosition
local loveAudioNewSource = love.audio.newSource
local loveAudioStop = love.audio.stop
local isFile = love.filesystem.isFile

local sounds = {}

local weakmeta = { __mode = "kv" }

local posX = 0
local posY = 0

local sfx = setmetatable({},weakmeta)
local sfxVolume = 1
local sfxLastId = 0

local music = setmetatable({},weakmeta)
local musicVolume = 1
local musicLastId = 0
local positionScale = 1/27

local musicCurrent
local musicCurrentId
local musicFading = {}

local sfxMeta = {
	play = function(s,x,y,v)
		if s.source:isPlaying() then
			s.clones[#s.clones+1] = s.source
			s.source = s.source:clone()
		end
		s.source:setPosition((x or posX)*positionScale,(y or posY)*positionScale,0)
		if v then s.volume = v; s.source:setVolume(v*sfxVolume) end
		s.source:play()
	end,
	setVolume = function(s,vol)
		s.volume = vol
		s.source:setVolume(vol*sfxVolume)
	end,
	stop = function(s)
		s.source:stop()
		for k,v in pairs(s.clones) do v:stop() end
	end,
	pause = function(s) s.source:pause() end,
	resume = function(s) s.source:resume() end,
	setPosition = function(s,x,y) s.source:setPosition(x*positionScale,y*positionScale) end,
	isPlaying = function(s) return s.source:isPlaying() end,
	isPaused = function(s) return s.source:isPaused() end,
	isStopped = function(s) return s.source:isStopped() end
}
sfxMeta.__index = sfxMeta

local musicMeta = {
	play = function(s)
		if musicCurrentId ~= s.id then
			if musicCurrent then
				musicFading[#musicFading+1] = {
					musicCurrent,
					musicCurrent:getVolume(), 0
				}
			end
			musicCurrentId = s.id
			musicCurrent = loveAudioNewSource(s.path,"stream")

			musicCurrent:setVolume(musicVolume*s.volume)

			if musicCurrent:getChannels() == 1 then
				musicCurrent:setRelative(true)
			end
			musicCurrent:setLooping(true)

			musicCurrent:play()
		end
	end,
	stop = function(s)
		if musicCurrentId == s.id and musicCurrent then
			musicFading[#musicFading+1] = {
				musicCurrent,
				musicCurrent:getVolume(), 0
			}

			musicCurrentId = nil
			musicCurrent = nil
		end
	end,
	pause = function(s)
		if musicCurrentId == s.id and musicCurrent then
			musicCurrent:pause()
		end
	end,
	resume = function(s)
		if musicCurrentId == s.id and musicCurrent then
			musicCurrent:resume()
		end
	end,
	rewind = function(s)
		if musicCurrentId == s.id and musicCurrent then
			musicCurrent:rewind()
		end
	end,
	seek = function(s,offset,unit)
		if musicCurrentId == s.id and musicCurrent then
			musicCurrent:seek(offset,unit)
		end
	end,
	tell = function(s,unit)
		if musicCurrentId == s.id and musicCurrent then
			musicCurrent:tell(unit)
		end
	end,
	setVolume = function(s,newVol)
		s.volume = newVol
		if musicCurrentId == s.id and musicCurrent then
			musicCurrent:setVolume(newVol*musicVolume)
		end
	end,
	isPlaying = function(s)
		if musicCurrentId == s.id and musicCurrent then
			return musicCurrent:isPlaying()
		else
			return false
		end
	end
}
musicMeta.__index = musicMeta

function sounds.setPositionScale(n)
	positionScale = 1 / n
end

function sounds.newSFX(path,volumeMod,id)
	local suc,new = pcall(loveAudioNewSource,path,"static")

	if suc then
		volumeMod = volumeMod or 1

		new:setVolume(volumeMod*sfxVolume)

		local obj = { source = new,volume = volumeMod,clones = setmetatable({},weakmeta) }
		setmetatable(obj,sfxMeta)

		if type(id) == "string" then
			sfx[id] = obj
			obj.id = id
		else
			sfxLastId = sfxLastId + 1
			sfx[sfxLastId] = obj
			obj.id = sfxLastId
		end

		return obj
	end
	log("Could not create SFX source... ["..path.."]")
end

local function newMusicObj(path,volumeMod,id)
	local obj = { path = path,volume = volumeMod or 1 }
	setmetatable(obj,musicMeta)

	if type(id) == "string" then
		music[id] = obj
		obj.id = id
	else
		musicLastId = musicLastId + 1
		music[musicLastId] = obj
		obj.id = musicLastId
	end

	return obj
end
function sounds.newMusic(...)
	if sounds.fileParser then
		local parse,source,volumeMod,id = ...
		local suc,d = pcall(sounds.fileParser,parse)

		if suc and d then
			local obj = newMusicObj(source,tonumber(d.volume) or volumeMod,d.id or id)

			for k,v in pairs(d) do
				if obj[k] == nil then
					obj[k] = v
				end
			end

			return obj
		else
			return newMusicObj(source,volumeMod,id)
		end
	else
		return newMusicObj(...)
	end
end

function sounds.setSFXVolume(newVolume)
	if type(newVolume) == "number" then
		sfxVolume = newVolume
		for k,v in pairs(sfx) do
			v.source:setVolume(v.volume*newVolume)
		end

		return true
	else
		return false
	end
end

function sounds.setMusicVolume(newVolume)
	if type(newVolume) == "number" then
		musicVolume = newVolume
		if musicCurrent then
			musicCurrent:setVolume(music[musicCurrentId].volume*newVolume)
		end

		return true
	else
		return false
	end
end

function sounds.setSFXPitch(pitch)
	for k,v in pairs(sfx) do
		v.source:setPitch(pitch)
		for k,v in pairs(v.clones) do
			v:setPitch(pitch)
		end
	end
end

function sounds.stopSFX()
	local musicP
	if musicCurrent then
		musicP = musicCurrent:tell()
	end
	loveAudioStop()
	if musicP then
		musicCurrent:play()
		musicCurrent:seek(musicP)
	end
end

function sounds.stopMusic()
	if musicCurrent then
		musicCurrent:stop()

		musicCurrent = nil
		musicCurrentId = nil
	end
end

function sounds.getCurrentMusic()
	return music[musicCurrentId],musicCurrent
end

function sounds.setListener(x,y)
	loveAudioSetPosition(x*positionScale,y*positionScale,-1.4)
	posX = x
	posY = y
end

local function update(dt)
	for k,v in pairs(musicFading) do
		v[3] = v[3] + dt
		if v[3] >= 1 then
			v[1]:stop()
			musicFading[k] = nil
		else
			v[1]:setVolume(v[2] * (1-v[3]))
		end
	end
end
sounds.update = update

sounds.sfxList = sfx
sounds.musicList = music
--[[sounds.fileParser = function(t)
	return love.filesystem.load(t)()
end]]

local threadedLoading = false
function sounds.enableThreadedLoading()
	if threadedLoading then
		log("Threaded Loading already enabled.")
	else
		log("Enabling Threaded Loading...")
		threadedLoading = true

		local waitingForSource = false
		local getChannel = love.thread.getChannel("sounds_library_loading_from")
		local toChannel = love.thread.getChannel("sounds_library_loading_to")

		musicMeta.play = function(s)
			if musicCurrentId ~= s.id then
				musicCurrentId = s.id
				waitingForSource = s
				toChannel:push(s.path)
			end
		end

		sounds.loadThread = love.thread.newThread[[
		local function log(...)
			print("[Sound Sec.]",...)
		end

		require "love.sound"
		local newAudio = require "love.audio" .newSource
		local getChannel = love.thread.getChannel("sounds_library_loading_from")
		local toChannel = love.thread.getChannel("sounds_library_loading_to")

		while true do
			local path = toChannel:demand()
			if path then
				log("Requested loading of ["..path.."]")
				if toChannel:getCount() >= 1 then
					for i=1,toChannel:getCount()-1 do toChannel:pop() end
					path = toChannel:pop()
					log("\tFound more requests, latest ["..path.."]")
				end
				local s,new = pcall(newAudio,path,"stream")
				log("\tLoaded file.")
				while toChannel:getCount() >= 1 do
					for i=1,toChannel:getCount()-1 do toChannel:pop() end
					path = toChannel:pop()
					s,new = pcall(newAudio,path,"stream")
					log("\tFound more requests, loaded ["..path.."]")
				end
				if s then
					getChannel:push(new)
					log("\tPushed source to main thread!")
				else
					getChannel:push("error")
					log("Could not load ["..path.."]",new)
				end
			end
		end
		]]
		sounds.loadThread:start()

		function sounds.update(dt)
			update(dt)

			if waitingForSource ~= nil then
				local source = getChannel:pop()
				if source ~= nil then
					if source ~= "error" then
						if musicCurrent then
							musicFading[#musicFading+1] = {
								musicCurrent,
								musicCurrent:getVolume(), 0
							}
						end

						musicCurrent = source

						musicCurrent:setVolume(musicVolume*waitingForSource.volume)

						if musicCurrent:getChannels() == 1 then
							musicCurrent:setRelative(true)
						end
						musicCurrent:setLooping(true)

						musicCurrent:play()
					end
				end
			end
		end
	end
end

return sounds
