local love = love
local GAME, INPUT, SOUNDS

require "errhand"

function love.load()
	print("Loading game...")
	love.keyboard.setTextInput(false)

	local arg = _arg

	-- Loading Libraries
	require "callbacks"
	colors  = require "libs.colors"
	input   = require "libs.input"
	anim    = require "libs.anim2"
	assets  = require "libs.assets"
	class   = require "libs.class"
	saveLib = require "libs.save"
	fileT   = require "libs.filetable"
	sounds  = require "libs.sounds"
	md5     = require "libs.md5"

	local input, anim, assets, class, saveLib, fileT, sounds, md5 = input, anim, assets, class, saveLib, fileT, sounds, md5
	local window, colors = window, colors

	random = assets.random("general")

	-- Rounding functions
	-- btw, rounding -0.5 to 0 is correct!
	local floor = math.floor
	function math.round(n)
		return floor(n + 0.5)
	end
	function math.roundn(n, f)
		return floor(n * f + 0.5) / f
	end

	--[[
	Configuration and setup.
	]]
	function cnfReset()
		cnf = {
			kl = "left", kr = "right",
			ku = "up", kd = "down",
			kj = "z", ka = "x",
			kp = "return", kc = "backspace",

			jl = "dpleft", jr = "dpright",
			ju = "dpup", jd = "dpdown",
			ju2 = "leftshoulder",
			jd2 = "rightshoulder",
			jj = "a", ja = "b",
			jp = "start", jc = "back",
			jt = 30,
			prompt = "ds4",

			fs = false,
			vsync = true,
			shadow = 0,
			pmax = 1000,

			sfx = 75,
			music = 25,

			char = "denniz",
			multi = false
		}
	end
	cnfReset()
	local cnf = cnf

	if love.filesystem.isFile("cnf") then
		local ncnf, err = saveLib.read("cnf")
		if ncnf then
			for k,v in pairs(ncnf) do
				if type(v) == type(cnf[k]) then
					cnf[k] = v
				end
			end
		else
			print("Could not load configuration:", err)
		end
	end

	-- Creating the window
	love.window.setMode(1280, 720, {
		fullscreen = cnf.fs,
		fullscreentype = "desktop",
		vsync = cnf.vsync,
		resizable = true,
		borderless = false,
		minwidth = 640,
		minheight = 360
	})
	love.window.setIcon(love.image.newImageData("assets/icon.png"))

	do
		-- Setting up Font and love.graphics
		local FONT = love.graphics.newFont("assets/font.ttf", 10)
		FONT:setFilter("linear", "nearest")
		local FONTADD = love.graphics.newFont("assets/font_add.ttf", 10)
		FONTADD:setFilter("linear", "nearest")

		FONT:setFallbacks(FONTADD)
		love.graphics.setFont(FONT)

		font = FONT

		love.graphics.setDefaultFilter("nearest", "nearest")
	end

	if not (love.filesystem.isFile("jnr/char/"..cnf.char..".lua") or love.filesystem.isFile("jnr/char/"..cnf.char..".dpp2c")) then
		cnf.char = "denniz"
	end

	do
		local who = random:random()
		if who < 0.475 then
			_protag = "denniz"
		elseif who < 0.95 then
			_protag = "diana"
		else
			_protag = cnf.char
		end
	end

	input.overrideCallbacks()

	input.bindKey(cnf.kl, "left")
	input.bindKey(cnf.kr, "right")
	input.bindKey(cnf.ku, "up")
	input.bindKey(cnf.kd, "down")
	input.bindKey(cnf.kj, "jump")
	input.bindKey(cnf.ka, "attack")
	input.bindKey(cnf.kp, "pause")
	input.bindKey(cnf.kc, "cselect")

	input.bindButton(cnf.jl, "left")
	input.bindButton(cnf.jr, "right")
	input.bindButton(cnf.ju, "up")
	input.bindButton(cnf.ju2,"up")
	input.bindButton(cnf.jd, "down")
	input.bindButton(cnf.jd2,"down")
	input.bindButton(cnf.jj, "jump")
	input.bindButton(cnf.ja, "attack")
	input.bindButton(cnf.jp, "pause")
	input.bindButton(cnf.jc, "cselect")

	input.bindAxis("leftx",  cnf.jt*0.01, "left", "right")
	input.bindAxis("lefty",  cnf.jt*0.01, "up", "down")
	input.bindAxis("righty", cnf.jt*0.01, "up", "down")

	local promptId = {
		keyboard = {
			left = "kl", right = "kr", up = "ku", down = "kd",
			jump = "kj", attack = "ka", pause = "kp"
		},
		gamepad = {
			left = "jl", right = "jr", up = "ju", down = "jd",
			jump = "jj", attack = "ja", pause = "jp"
		}
	}

	function input.getPrompt(id)
		if input.getLastType() == "keyboard" then
			return input.getKeyName(cnf[ promptId.keyboard[id] ])
		else
			return input.getButtonName(cnf[ promptId.gamepad[id] ], cnf.prompt)
		end
	end

	sounds.setSFXVolume(cnf.sfx*0.01)
	sounds.setMusicVolume(cnf.music*0.01)

	if cnf.multi then
		input.separate(true)
	end

	unlocks = {}
	if love.filesystem.isFile("unlocks") then
		local nunlocks, err = saveLib.read("unlocks")
		if nunlocks then
			for k,v in pairs(nunlocks) do
				unlocks[k] = v
			end
		else
			print("Could not load unlocks:", err)
		end
	end

	_secrets = {}

	--[[
	Hard-Coded key events (F-keys)
	]]
	input.bindKey("f12", "screenshot")
	input.bindEvent(function()
		local sc = love.graphics.newScreenshot(false)
		if not love.filesystem.isDirectory("screenshots") then
			love.filesystem.createDirectory("screenshots")
		end

		local date = os.date("%y-%m-%d_%H.%M.%S")
		local try = 0
		local filename = "screenshots/"..date..".png"
		while love.filesystem.isFile(filename) do
			try = try + 1
			filename = "screenshots/"..date.."-"..tostring(try)..".png"
		end
		sc:encode("png", filename)

		GAME.message.queue({
			colors.white, "Took Screenshot!\nSaved as ",
			colors.yellow, "'"..filename:gsub("^screenshots/", "").."'",
			colors.white, "!"}, 1)
	end, "screenshot")

	input.bindKey("f2", "zoomOut")
	input.bindEvent(function()
		if window.sM > 1 then
			window.sM = window.sM * (2/3)
			love.resize(window.w, window.h)
		end
	end, "zoomOut")

	input.bindKey("f3", "zoomIn")
	input.bindEvent(function()
		if window.sM < (3^3/2^3) then
			window.sM = window.sM * (3/2)
			love.resize(window.w, window.h)
		end
	end, "zoomIn")

	-- Toggling Fullscreen
	local function toggleFullscreen()
		if love.window.getFullscreen() then
			love.window.setFullscreen(false)
			cnf.fs = false
		else
			love.window.setFullscreen(true, "desktop")
			cnf.fs = true
		end
	end
	window.toggleFullscreen = toggleFullscreen

	input.bindKey("f11", "fullscreen")
	input.bindEvent(toggleFullscreen, "fullscreen")

	--if cnf.fs then toggleFullscreen() end

	-- Loading all sounds the game will need very often
	sfx = {
		--jump    = sounds.newSFX("assets/sfx/jump.ogg"   , 0.60, "jump"),
		colide  = sounds.newSFX("assets/sfx/colide.ogg" , 1.00, "colide"),
		death   = sounds.newSFX("assets/sfx/death.ogg"  , 1.00, "death"),
		collect = sounds.newSFX("assets/sfx/collect.ogg", 0.60, "collect"),
		swing   = sounds.newSFX("assets/sfx/swing.ogg"  , 0.45, "swing"),
		smalexp = sounds.newSFX("assets/sfx/smalexp.ogg", 0.80, "smalexp"),
		explode = sounds.newSFX("assets/sfx/explode.ogg", 1.00, "explode"),
		select  = sounds.newSFX("assets/sfx/select.ogg" , 1.00, "select"),
	}
	-- Creating Music Objects (doesn't actually load the mp3)
	sounds.fileParser = fileT.load
	do
		local MUSIC = {}
		for k,v in ipairs(love.filesystem.getDirectoryItems("assets/music")) do
			if v:find("%.mp3$") then
				local raw = v:gsub("%.mp3$", "")
				MUSIC[raw] = sounds.newMusic("assets/music/"..raw..".meta", "assets/music/"..raw..".mp3", 0.13, raw)
			end
		end

		function MUSIC.set(id)
			if MUSIC.current ~= MUSIC[id] then
				if MUSIC[id] == nil then
					MUSIC.current:stop()
				else
					if cnf.music > 0 then
						MUSIC[id]:play()

						GAME.message.queue({
							colors.softBlue, "♫ Now playing ♫\n",
							colors.white, tostring(MUSIC[id].full_title or "A track"),
							colors.softBlue, "\nby ",
							colors.white, tostring(MUSIC[id].author or "unknown")
						}, 2)
					end
				end

				MUSIC.current = MUSIC[id]
			end
		end

		music = MUSIC
	end
	-- And this thing...
	sounds.enableThreadedLoading()

	do
		-- Managing/Resetting Save Data
		local SAVE = {}
		save = SAVE

		local currentSave
		function SAVE.write()
			if currentSave then
				saveLib.write("save"..tostring(currentSave), SAVE.data)
			end
		end

		function SAVE.load(n)
			local file = "save"..tostring(n or currentSave)

			SAVE.reset()
			jnr.inHub = true

			if love.filesystem.isFile(file) then
				local load = saveLib.read(file)
				for k,v in pairs(load) do
					if type(v) == type(SAVE.data[k]) then
						SAVE.data[k] = v
					end
				end

				SAVE.data.totalCollect = 0
				for k,v in pairs(SAVE.data.collect) do
					SAVE.data.totalCollect = SAVE.data.totalCollect + v
				end
			end
		end

		function SAVE.set(n)
			currentSave = n
		end

		function SAVE.get()
			return currentSave
		end

		function SAVE.reset()
			SAVE.data = {
				hub = "",
				coord = {},
				flags = {},
				collect = {},
				totalCollect = 0,
				deaths = 0,
				time = 0,
				hardcore = false
			}
		end
		SAVE.reset()

		local startArea = "tutorial"
		function SAVE.init(slot, hardcore, area)
			SAVE.set(slot)
			SAVE.load()
			if hardcore ~= nil then
				SAVE.data.hardcore = hardcore
			end

			game.queueTask(jnr.resetAll)
			game.setState("jnr", area or SAVE.data.hub == "" and startArea or SAVE.data.hub)
		end

		local floor = math.floor
		function SAVE.formatTime(time)
			time = floor(time)
			local sec = time%60

			time = floor(time*(1/60))
			local min = time%60

			time = floor(time*(1/60))
			local hour = time

			return ("%02d:%02d:%02d"):format(hour, min, sec)
		end

		function SAVE.unlock(id, type)
			if not unlocks[id] then
				unlocks[id] = true

				if type == "char" then
					game.message.queue({
						colors.white, "Unlocked a new character!\n",
						colors.gold, fileT.get("jnr/char/#names.txt", id) or id,
						colors.white, "\nSelect them in the Settings!"
					})
				elseif type == "secret" then
					game.message.queue({
						colors.white, "New secret discovered!\n",
						colors.gold, fileT.get("menu/data/secrets/#names.txt", id) or id,
						colors.white, "\nSelect them in the Bonus menu!"
					})
				end

				saveLib.write("unlocks", unlocks)
				return true
			else
				return false
			end
		end

		function SAVE.deunlock(id)
			unlocks[id] = nil

			saveLib.write("unlocks", unlocks)
		end
	end

	newLog = assets("logging", function(module)
		return function(...)
			print(module, ...)
		end
	end)

	-- Game State Handler
	GAME = require "game"
	game = GAME
	-- All States
	jnr = require "jnr"
	menu = require "menu"
	scene = require "scene"
	title = require "title"

	mods = require "mods"

	GAME.setState("title", "menu", "main", "main")

	do
		-- Setting a splash
		local allSplashes = {}
		for line in love.filesystem.lines("assets/splashes.txt") do
			allSplashes[#allSplashes+1] = line
		end
		local splash = allSplashes[random:random(1, #allSplashes)]
		love.window.setTitle(_gameTitle.."  ––  "..splash)
		_gameSplash = splash

		if splash == "Press F to pay respect." then
			if not input.isKeyBound("f") then
				input.bindKey("f", "EEpay_respect")
				input.bindEvent(function()
					game.message.queue("Payed respect.\nHave a nice day.")
					input.bindKey("f", nil)
				end, "EEpay_respect")
			end
		end
	end

	if arg.secrets then
		-- Enabling secrets via command line
		for s in (tostring(arg.secrets)..","):gmatch("([^,]+),") do
			if unlocks[s] and love.filesystem.isFile("menu/data/secrets/"..s..".lua") then
				local this = require("menu.data.secrets."..s)
				if type(this) == "function" then
					this()
				end
				_secrets[s] = this
			end
		end
		game.message.clear()
	end

	if love.filesystem.isFile(".version") then
		local lastVersion = love.filesystem.read(".version")
		if lastVersion ~= _gameVersion then
			print("Game version mismatch!")

			local a, b, c = _gameVersion:match("^(%d+)%.(%d+)%.(%d+)$")
			a, b, c = tonumber(a), tonumber(b), tonumber(c)

			local d, e, f = lastVersion:match("^(%d+)%.(%d+)%.(%d+)$")
			if d and e and f then
				d, e, f = tonumber(d), tonumber(e), tonumber(f)

				if a < d or (a == d and b < e) or (a == d and b == e and c < f) then
					print("\tOlder version!?")
					-- Running an outdated version again...
					game.message.queue{
						colors.yellow, "You are running a version",
						colors.red, " older ",
						colors.yellow, "than the one that has last been run!\nTo prevent data loss, unexpected behavior, crashes or more, it is recommended to",
						colors.red, " close ",
						colors.yellow, "the game and",
						colors.red, " update to/run the current version",
						colors.yellow, "."
					}
				else
					print("\tNewer version!")
					-- Running a newer version for the first time.
					-- Compatibility if the need arises.
					love.filesystem.write(".version", _gameVersion)
				end
			else
				print("\tLast version string invalid...")
				-- Version string invalid...?
				love.filesystem.write(".version", _gameVersion)
			end
		end
	else
		print("First time running! Thank you for playing DPlox Parkour 2! I hope you have a nice day and enjoy your stay!")
		-- Running the game for the first time at all.
		love.filesystem.write(".version", _gameVersion)

		menu.init("main", "main")
		GAME.setState("title", "menu", "options", "main")

		game.message.queue "This seems to be your first time running this game.\nHow about you get nice and cozy and set up everything to your liking?\nSounds good? Okay. Take your time.\nJust hit 'Back' when you're finished."
	end

	love.mouse.setVisible(false)

	debugging = false
	if arg.debug then
		require "load_debug"
	end

	love.resize(love.graphics.getDimensions())
	INPUT, SOUNDS = input, sounds

	collectgarbage()
end

function love.update(dt)
	if dt > 0.05 then dt = 0.05 end

	GAME.update(dt)
	INPUT.update(dt)
	SOUNDS.update(dt)
end

function love.draw(dt)
	GAME.draw(dt)
end

function love.run()
	love.math.setRandomSeed(os.time())

	love.load()

	-- We don't want the first frame's dt to include time taken by love.load.
	love.timer.step()

	local dt = 0

	-- Main loop time.
	while true do
		-- Process events.
		love.event.pump()
		for name, a,b,c,d,e,f in love.event.poll() do
			if name == "quit" then
				if not love.quit() then
					return a
				end
			end
			love.handlers[name](a,b,c,d,e,f)
		end

		-- Update dt
		love.timer.step()
		dt = love.timer.getDelta()

		-- Call update and draw
		love.update(dt)

		if love.graphics.isActive() then
			love.graphics.clear()
			love.graphics.origin()
			love.draw(dt)
			love.graphics.present()
		end

		love.timer.sleep(0.001)
	end
end
