local love = love
local floor, abs, pi, huge = math.floor, math.abs, math.pi, math.huge
local window, input, colors = window, input, colors
local font, sfx = font, sfx

local log = newLog "[Menu]"

local menu = {}

local data, curr, prev, substate
local select, cancel, label, allText
local graphic, color, bgColor
local length, spacing, thin
local jnrData

local defaultColor = colors.new(0, 127, 255)

function menu.deinit()
	log("Clearing temporary data...")

	data, curr, prev, substate = nil
	select, cancel, label, allText = nil
	color, bgColor = nil
	length, spacing, thin = nil
	jnrData = nil
	menu.data, menu.substate = nil
end

function menu.init(id, nsubstate, back)
	log("Initializing...")

	color = defaultColor

	local lastData
	if data then
		lastData = data
		if not back then lastData.select = select end
	end

	if not allText then allText = love.graphics.newText(font) end

	select = 1
	cancel = true

	if back and data and id == data.prev then
		-- Going back
		data = data.prevData
	elseif love.filesystem.isFile("menu/data/"..id) then
		log("\tConverting content to raw data...")
		data = {}
		for line in love.filesystem.lines("menu/data/"..id) do
			local sub
			if line == "" then
				sub = false
			elseif line:find("^;;") then
				local arg, value = line:match("^;;(.-)=(.*)$")
				if arg == "select" then
					value = tonumber(value)
					if value then
						data.select  = value
					end
				elseif arg == "cancel" then
					if value == "false" or value == "nil" then
						data.cancel = false
					elseif value:find("^%[.*%]$") then
						local s, r = pcall(loadstring, value:sub(2, #value-1))
						if not s then log("\t\tError loading Lua code for cancelling: ", r) end
						data.cancel = r
					elseif tonumber(cancel) then
						data.cancel = tonumber(cancel)
					else
						data.cancel = true
					end
				elseif arg == "color" then
					local nc = { value:match("^(%d+),(%d+),(%d+)$") }
					for i=1, #nc do
						nc[i] = tonumber(nc[i]) or 255
					end
					data.color = colors.new(nc[1], nc[2], nc[3], nc[4])
				elseif arg == "run" then
					local s, r = pcall(loadstring(value:sub(2, #value-1)), data)
					if not s then log("\t\tError running code: ", r) end
				else
					data[arg] = value
				end
			else
				local text, desc, action = line:match("^(.-);;(.-);;(.*)$")
				if not (text and action and desc) then
					text, action = line:match("^(.-);;(.*)$")
					if not (text and action) then
						text = line
						action = nil
					end
				else
					desc = desc:gsub("\\n", "\n")
				end

				if action then
					if action == "" then
						action = nil
					elseif action:find("^%[.*%]$") then
						local s, r = pcall(loadstring, action:sub(2, #action-1))
						if not s then log("\t\tError loading Lua code for '"..text.."': ", r) end
						action = r
					end
				end

				if text:find("%%.-%%") then
					local from = text
					local to = {colors.white}
					local i = 1
					while true do
						local a, b = from:find("%%.-%%")
						if a and b then
							to[#to+1] = from:sub(1, a-1)
							to[#to+1] = colors[from:sub(a+1, b-1)] or colors.white
							from = from:sub(b+1, #from)
						else
							to[#to+1] = from
							break
						end
					end
					text = to
				end

				sub = {
					text = text,
					desc = desc,
					action = action
				}
			end
			data[#data+1] = sub
		end
	elseif love.filesystem.isFile("menu/data/"..id..".lua") then
		log("\tLoading raw data...")
		data = love.filesystem.load("menu/data/"..id..".lua")()
	else
		error("Please the check the menu data ID and/or file extension!")
	end

	if data.select ~= nil then
		select = data.select
	end

	for i=1, #data do local v = data[i] if v then
		if i == select then v.offset = 16 else v.offset = 0 end
		if v.action == nil and (v.actionR ~= nil or v.actionL ~= nil) then v.action = v.actionR or v.actionL end
		if v.text then v.width = font:getWrap(v.text, huge) v.__text = v.text end
	end end

	if data.cancel ~= nil then cancel = data.cancel end
	if data.color  ~= nil then color  = data.color end
	if data.label  ~= nil then label  = data.label else label = "" end
	if not back then data.prev = curr data.prevData = lastData end

	length = #data

	if not graphic then
		graphic = {}
		local image = assets.image("assets/menu.png")
		graphic.image = image
		graphic.batch = love.graphics.newSpriteBatch(image, 20, "stream")

		graphic.thick = love.graphics.newQuad(0,  0, 196, 14, 256, 256)
		graphic.thin  = love.graphics.newQuad(0, 14, 196, 13, 256, 256)
		graphic.logo  = love.graphics.newQuad(0, 27, 256, 64, 256, 256)
	end

	if data.thin then
		spacing = 13
		thin = true
		graphic.button = graphic.thin
	else
		spacing = 16
		thin = false
		graphic.button = graphic.thick
	end

	graphic.batch:setColor(color)

	if nsubstate == "main" then
		bgColor = color * colors.vec4(0.35, 0.35, 0.35, 1)
	elseif nsubstate == "pause" then
		bgColor = color * colors.vec4(0.35, 0.35, 0.35, 0.5)
	end

	if nsubstate == "main" then
		if substate ~= "main" then
			log("\tLoading player...")
			log("\t\tLoading level data...")
			local leveldata = love.filesystem.load("jnr/level/main_menu.lua")()

			local tiles = jnr.Tiles(leveldata)

			local player = jnr.Player(leveldata.spawn[1], leveldata.spawn[2], 0, jnr.getPlayerData(leveldata.char))
			tiles:addObject(player)

			local particle = jnr.particle
			particle:clear()

			tiles.particle = particle
			particle.tiles = tiles

			tiles:backup()

			jnrData = {
				player = player,
				tiles = tiles,
				particle = particle
			}
		end
		sounds.setListener(900, 0)
	else
		jnrData = nil
	end

	substate = nsubstate

	prev = curr
	curr = id

	menu.data, menu.substate = data, substate
end

function menu.update(dt)
	if input.rep("down") then
		select = select + 1
		while not data[select] do
			if select > length then
				select = 1
			else
				select = select + 1
			end
		end
		sfx.select:play()
	elseif input.rep("up") then
		select = select - 1
		while not data[select] do
			if select < 1 then
				select = length
			else
				select = select - 1
			end
		end
		sfx.select:play()
	end

	if input.new("jump") then
		local button = data[select]
		if button then
			local at = type(button.action)
			if at == "function" then
				button:action()
				input.releaseInput("jump")
				sfx.select:play()
				return
			elseif at == "string" then
				if button.action == "!" then
					game.setState("menu", data.prev, substate, true)
				else
					game.setState("menu", button.action, substate)
				end
				input.releaseInput("jump")
				sfx.select:play()
				return
			end
		end
	elseif input.rep("right") then
		local button = data[select]
		if button and button.actionR then
			button:actionR()
			sfx.select:play()
		end
	elseif input.rep("left") then
		local button = data[select]
		if button and button.actionL then
			button:actionL()
			sfx.select:play()
		end
	elseif input.new("attack") then
		if cancel then
			local ct = type(cancel)
			if ct == "function" then
				cancel()
			elseif ct == "number" then
				local button = data[cancel]
				local bt = type(button)
				if bt == "function" then
					button:action()
				elseif bt == "string" then
					if bt == "!" then
						game.setState("menu", data.prev, substate, true)
						data.prev = nil
					else
						game.setState("menu", button, substate)
					end
				end
			else
				game.setState("menu", data.prev, substate, true)
			end
			input.releaseInput("attack")
			sfx.select:play()
			return
		end
	end

	for i=1, length do local v = data[i] if v then
		if i == select then
			v.offset = 16 - (16 - v.offset) * 0.000025 ^ dt
		else
			v.offset = v.offset * 0.000025 ^ dt
		end
	end end

	if data.update then data:update(dt, select) end

	if jnrData then
		local player = jnrData.player
		player:update(dt)
		if player.x + player.w + 6 < 900-window.vw*0.5 then
			player.x = 905+window.vw*0.5
		elseif player.x - 6 > 900+window.vw*0.5 then
			player.x = 895-window.vw*0.5
		end

		jnrData.particle:update(dt)
	end
end

local promptText = {
	colors.white, "",
	colors.white, "",
	colors.white, "",
}

function menu.draw()
	allText:clear()

	if substate == "main" then
		love.graphics.clear(bgColor)
	elseif substate == "pause" then
		jnr.draw()
		love.graphics.setColor(bgColor)
		love.graphics.rectangle("fill", 0, 0, window.w, window.h)
		love.graphics.setColor(colors.white)

		local slot = save.get()
		allText:addf(slot and "Slot "..slot or "No-Save Slot", window.vw, "center", 0, 0)
	end

	love.graphics.push()

	love.graphics.scale(window.s)

	local xc = floor(window.vw*0.5)
	local by = floor(window.vh*0.5 - length*spacing*0.5 + 1)
	local y = by

	graphic.batch:clear()

	if curr == "main" then
		graphic.batch:setColor(colors.white)
		graphic.batch:add(graphic.logo, xc-128, 1)
	end

	graphic.batch:setColor(color)
	for i=1, length do
		local v = data[i]
		if v then
			local x = xc - 106 + v.offset
			graphic.batch:add(graphic.button, x, y)

			if v.text ~= v.__text then v.width = font:getWrap(v.text, huge) v.__text = v.text end
			if v.width <= 166 then
				allText:add(v.text, x + 15, y + 1)
			else
				allText:add(v.text, x + 15, y + 1, 0, 166/v.width, 1)
			end
		end
		y = y + spacing
	end

	love.graphics.draw(graphic.batch)

	love.graphics.setColor(colors.black(1, 1, 1, 0.5))
	for i=1, length do
		local v = data[i]
		if v then
			if v.offset > 2 and v.desc then
				local xw = floor(window.vw*0.8)

				local w, wrap = font:getWrap(v.desc, xw)
				local h = #wrap * 10

				love.graphics.rectangle("fill", 0, window.vh+64-h-v.offset*5, window.vw, h+8)

				allText:addf(v.desc, xw, "center", floor(window.vw*0.1), window.vh+68-h-v.offset*5)
			end
		end
	end
	love.graphics.setColor(colors.white)

	allText:addf(label, window.vw, "right", 0, 0)

	local button = data[select]
	if button and (button.actionR or button.actionL) then
		promptText[2] = "Select> "..input.getPrompt("up").." | "..input.getPrompt("down").." | "..input.getPrompt("left").." | "..input.getPrompt("right").."\n"
	else
		promptText[2] = "Select> "..input.getPrompt("up").." | "..input.getPrompt("down").."\n"
	end
	if button and button.action then
		promptText[4] = "Accept> "..input.getPrompt("jump").."\n"
	else
		promptText[4] = ""
	end
	if cancel then
		promptText[6] = "Back> "..input.getPrompt("attack")
	else
		promptText[6] = ""
	end
	allText:add(promptText, 1, 0)

	love.graphics.setColor(colors.black)
	love.graphics.draw(allText, 1, 1)
	love.graphics.setColor(colors.white)
	love.graphics.draw(allText, 0, 0)

	if data.draw then data:draw(by, select) end

	if jnrData then
		love.graphics.translate(floor(window.vw*0.5-900), by)

		jnrData.player:draw(window.s)
		jnrData.particle:draw()
	end

	love.graphics.pop()
end

return menu
