local love = love
local floor, ceil, round, roundn, max = math.floor, math.ceil, math.round, math.roundn, math.max
local pairs, pcall, type, loadstring, tostring, tonumber = pairs, pcall, type, loadstring, tostring, tonumber
local window, input, sounds, colors = window, input, sounds, colors
local random = random

local log = newLog "[Scene]"

local scene = {}

local Actor = require "scene.actor"
local SlowText = require "scene.slow_text"

scene.Actor = Actor
scene.SlowText = SlowText

local script, progress, actors, hud
local waiting, currentText, currentEvent, currentSpeaker
local tiles, scroll, resetFinish

local shake = {
	power = 0,
	x = 0, y = 0, r = 0
}

function scene.init(new, arg)
	log("Setting new scene...!")

	tiles = jnr.tiles
	tiles:backup()

	scroll = {
		toX = jnr.scroll.toX, toY = jnr.scroll.toY,
		y = jnr.scroll.y, x = jnr.scroll.x
	}

	local removeObj = true
	resetFinish     = true

	if arg then
		if arg.removeObj   ~= nil then removeObj   = arg.removeObj   end
		if arg.resetFinish ~= nil then resetFinish = arg.resetFinish end
	end

	if removeObj or removeObj == nil then
		tiles.object = {}
	end

	script = {}

	local function loadArg(str)
		return loadstring("return {"..str.."}")()
	end

	log("\tLoading script...")
	local raw = love.filesystem.read("scene/data/"..new..".dpp2s")
	for line in (raw.."\n"):gmatch("%s*(..-)[\r\n]") do
		if not line:find("^/") then
			-- Script line format:
			-- TYPE ARG1, ARG2, ...
			local event = {}

			event.type = line:match("^(%S+)")
			local args = line:match("^%S+%s+(.*)$")

			if args then
				local s
				s, args = pcall(loadArg, args)
				if s then
					event.args = args
				else
					event.args = {}
					log("\t\tError in script line '"..tostring(#script+1).."':", args)
				end
			else
				event.args = {}
			end
			script[#script+1] = event
		end
	end

	log("\tDone with parsing!")

	currentEvent = {
		type = "none", args = {}
	}
	currentText, currentSpeaker = nil
	actors = {}

	progress = 0
	waiting = false

	hud = {}
	hud.image = assets.image("assets/scene.png")
	hud.box = love.graphics.newQuad(0, 0, 256, 48, 256, 256)

	scene.progressScript(0, false)
	scroll.x = scroll.toX
	scroll.y = scroll.toY

	shake.power = 0
	scene.script, scene.actors, scene.scroll = script, actors, scroll
	log(">> Set Scene!")
end

function scene.deinit()
	tiles:resetNow()

	currentText, currentSpeaker = nil
	script, actors = nil
	scene.script, scene.actors = nil
end

function scene.progressScript(dt, allowReturn)
	if waiting then
		local waitingT = type(waiting)
		if waitingT == "number" then
			waiting = waiting - dt
			if waiting <= 0 then
				waiting = false
			end
		elseif waitingT == "function" then
			if waiting() then
				waiting = false
			end
		elseif currentEvent.type == "text" then
			if currentText and input.new("jump") then
				if not currentText.terminated then
					currentText:update(99)
				else
					waiting = false
				end
			end
		else
			waiting = false
		end
	end

	if currentText then
		currentText:update(dt)
	end

	while not waiting do
		progress = progress + 1

		local event = script[progress]
		if event then
			local kind, args = event.type, event.args
			local thisActor = actors[args[1]]

			if kind == "text" or kind == "text2" then
				if #args == 2 and type(args[2]) == "number" then
					currentText = SlowText(args[1], args[2])
				elseif #args == 1 then
					currentText = SlowText(args[1])
				else
					currentText = SlowText(args)
				end
				currentText.speaker = currentSpeaker
				if kind == "text" then waiting = true end
			elseif kind == "cleartext" then
				currentText = nil
			elseif kind == "newactor" then
				actors[args[1]] = tiles:addObject(Actor(args[2], args[3], love.filesystem.load("scene/actors/"..args[1]..".lua")()))
			elseif kind == "getobject" then
				local matching
				for key,obj in pairs(tiles.object) do
					local match = true
					for k,v in pairs(args[2]) do
						if type(k) == "table" and type(obj[k]) == "table" then
							for l,w in pairs(v) do
								if obj[k][l] ~= w then
									match = false
									break
								end
							end
							if not match then break end
						elseif obj[k] ~= v then
							match = false
							break
						end
					end
					if match then
						matching = obj
						break
					end
				end

				if matching then
					actors[args[1]] = matching
				else
					log("Could not find object...")
				end
			elseif kind == "move" then
				if thisActor then
					thisActor:moveActor(args[2], args[3])
				end
			elseif kind == "jump" then
				if thisActor then
					thisActor:jumpActor(args[2])
				end
			elseif kind == "setv" then
				if thisActor then
					thisActor.xv = args[2]
					thisActor.yv = args[3]
				end
			elseif kind == "anim" then
				if thisActor then
					thisActor:fixAnim(args[2])
				end
			elseif kind == "key" then
				if thisActor then
					thisActor[args[2]] = args[3]
				end
			elseif kind == "speaker" then
				if args[1] == nil then
					currentSpeaker = nil
				elseif thisActor then
					currentSpeaker = thisActor
				end
			elseif kind == "scroll" then
				if args[1] == nil then
					scroll.follow = nil
				elseif thisActor then
					scroll.follow = thisActor
				else
					scroll.toX = args[1] * tiles.ts-tiles.ts*0.5
					scroll.toY = args[2] * tiles.ts-tiles.ts*0.5
				end
			elseif kind == "shake" then
				shake.power = shake.power + (args[1] or 1)
			elseif kind == "run" then
				log("Running code...", pcall(args[1]))
			elseif kind == "runactor" then
				if thisActor then
					log("Running code on actor...", pcall(args[2], thisActor))
				end
			elseif kind == "message" then
				game.message.queue(args[1])
			elseif kind == "break" then
				waiting = true
			elseif kind == "wait" then
				if type(args[1]) == "string" then
					-- Parsing a string into a function
					if args[1]:find("^actor:") then
						local check
						local who, key, subkey, val = args[1]:match("^actor:([^%.]*)%.([^%.]*)%.([^=]*)==(.*)$")
						if subkey then
							check = actors[who] and actors[who][key] or nil
							key = subkey
						else
							who, key, val = args[1]:match("^actor:([^%.]*)%.([^=]*)==(.*)$")
							check = actors[who] or nil
						end
						if check ~= nil and val then
							if     val == "true"  then val = true
							elseif val == "false" then val = false
							elseif val == "nil"   then val = nil
							elseif tonumber(val)  then val = tonumber(val) end
							if check[key] == val then
								-- Condition already fulfilled
								waiting = true
								log("Special waiting skipped (condition met).")
							else
								-- Otherwise
								function waiting()
									if check[key] == val then
										waiting = false
									end
								end
								log("Special waiting function created.")
							end
						end
					end
				elseif args[1] then
					waiting = args[1]
				elseif currentText then
					waiting = currentText.length * max(currentText.speed, 0.05) + 1
				end
			end

			currentEvent = event
		else
			if allowReturn then
				log("Finished scene...! Returning to 'jnr'...")
				game.setState("jnr", resetFinish and jnr.reset or game.noInit)
				break
			else
				return true
			end
		end
	end
end

function scene.update(dt)
	if input.new("pause") then
		log("Skipping scene... Returning to 'jnr'...")
		game.setState("jnr", resetFinish and jnr.reset or game.noInit)
	else
		scene.progressScript(dt, true)
	end

	tiles:update(dt)

	if scroll.follow then
		scroll.toX = scroll.follow.x
		scroll.toY = scroll.follow.y
	end
	scroll.x = scroll.x + 1.5 * window.sM * dt * (scroll.toX - scroll.x)
	scroll.y = scroll.y + 1.5 * window.sM * dt * (scroll.toY - scroll.y)

	if shake.power > 0 then
		shake.x = (random:random()*2-1)*shake.power
		shake.y = (random:random()*4-2)*shake.power
		shake.r = (random:random()*0.1-0.05)*shake.power

		shake.power = shake.power - dt
		if shake.power < 0 then shake.power = 0 end
	end

	sounds.setListener(scroll.x, scroll.y)
end

function scene.draw()
	love.graphics.push()

	love.graphics.scale(window.s)

	-- Basically drawing the "JNR section"...
	local lx = roundn(-scroll.x + window.vw*0.5, window.s)
	local ly = roundn(-scroll.y + window.vh*0.5, window.s)
	tiles:drawBackground(lx, ly, window.vw, window.vh)

	if shake.power > 0 then
		local s = (shake.power*0.05+1) * window.s
		love.graphics.scale(shake.power*0.05+1)
		love.graphics.translate(roundn(window.w/s*0.5, s), roundn(window.h/s*0.5, s))
		love.graphics.rotate(shake.r)
		love.graphics.translate(roundn(-scroll.x + shake.x, s), roundn(-scroll.y + shake.y, s))
	else
		love.graphics.translate(lx, ly)
	end

	tiles:draw(-lx, -ly, window.vw, window.vh, window.s)

	love.graphics.translate(-lx, -ly)

	scene.drawBox(lx, ly)

	love.graphics.pop()
end

function scene.drawBox(lx, ly)
	if currentText then
		local s = round(max(window.w / 426, window.h / 240) * window.sM) / window.s

		-- Drawing the text box...
		local x = floor(window.vw*0.5-128*s)
		local y = floor(window.vh-48*s)

		love.graphics.draw(hud.image, hud.box, x, y, 0, s)

		love.graphics.setColor(colors.black)
		if currentText.speaker then
			local speak = currentText.speaker
			love.graphics.polygon("fill", x+82, y, lx+speak.x+speak.w*0.5, ly+speak.y+speak.h, x+104, y)
		end

		love.graphics.printf(currentText.print, x+4+s, y+4+s, 248, "left", 0, s)
		love.graphics.setColor(colors.white)
		love.graphics.printf(currentText.print, x+4, y+4, 248, "left", 0, s)

		if currentText.terminated and currentEvent.type == "text" then
			local text = "Continue> "..input.getPrompt("jump")
			love.graphics.setColor(colors.black(1, 1, 1, currentText.terminated))
			love.graphics.printf(text, x+6, window.vh-11*s, 248, "right", 0, s)
			love.graphics.setColor(colors.yellow(1, 1, 1, currentText.terminated))
			love.graphics.printf(text, x+4, window.vh-12*s, 248, "right", 0, s)
			love.graphics.setColor(colors.white)
		end
	end
end

return scene
